import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, Pango

from datetime import datetime
from ks_includes.screen_panel import ScreenPanel
from ks_includes.widgets.keypad import Keypad


class Panel(ScreenPanel):
    def __init__(self, screen, title, extra=None):
        title = _("Spool editor")
        super().__init__(screen, title)
        self.selected_weight_mode = "filament"
        self.saved_weight = 0
        self.numpad_visible = False
        self._measurement_buttons = {}
        self._progress_bar = None
        self.grid = Gtk.Grid(row_homogeneous=True, column_homogeneous=True, hexpand=True, vexpand=True)
        self.set_extra(extra)

    def set_extra(self, extra):
        self.spool = extra
        for child in self.content.get_children():
            self.content.remove(child)
        self._build_grid()
        self.content.add(self.grid)
        self.content.show_all()

    def _build_grid(self):
        self.grid = Gtk.Grid(row_homogeneous=True, column_homogeneous=True, hexpand=True, vexpand=True)

        if self.spool is None:
            box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, vexpand=True)
            label = Gtk.Label(label=_("No spool selected"), wrap=True, wrap_mode=Pango.WrapMode.WORD_CHAR)
            box.set_valign(Gtk.Align.CENTER)
            box.add(label)
            self.grid.attach(box, 0, 0, 1, 1)
            return

        if self._screen.vertical_mode:
            self.grid.attach(self._build_first_panel(), 0, 0, 1, 2)
            self.grid.attach(self._build_second_panel(), 0, 2, 1, 2)
        else:
            self.grid.attach(self._build_first_panel(), 0, 0, 1, 1)
            self.grid.attach(self._build_second_panel(), 1, 0, 1, 1)

    def _build_first_panel(self):
        title_box = Gtk.Box(hexpand=True)
        title_box.get_style_context().add_class("spool_title")

        icon_pixbuf = self.spool.icon
        icon = Gtk.Image.new_from_pixbuf(icon_pixbuf)
        icon.set_halign(Gtk.Align.CENTER)
        icon.get_style_context().add_class("spool_icon")

        title_box.pack_start(icon, False, False, 0)
        title = Gtk.Label(halign=Gtk.Align.START, wrap=True, wrap_mode=Pango.WrapMode.WORD_CHAR)
        title.set_markup(f"<big><b>{self._get_display_name()}</b></big>")
        title_box.pack_start(title, False, False, 0)

        buttons = self._build_measurement_details_box()

        info = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=10, hexpand=True, vexpand=True)
        info.set_valign(Gtk.Align.FILL)
        info.pack_start(title_box, False, False, 0)
        info.pack_start(buttons, True, True, 0)

        return info

    def _build_measurement_details_box(self):
        buttons = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=4, hexpand=True, vexpand=True)
        buttons.set_margin_start(24)
        buttons.set_margin_end(12)

        self._measurement_buttons = {}
        for i, row in enumerate(self._get_measurement_rows()):
            btn = self._gtk.Button(
                label=self._format_measurement_row_markup(row),
                style=f"color{i % 4 + 1}",
            )
            btn.connect("clicked", self._show_numpad, row["mode"])
            self._measurement_buttons[row["mode"]] = btn
            buttons.pack_start(btn, True, True, 0)

        return buttons

    def _get_measurement_rows(self):
        rows = []
        remaining_weight = getattr(self.spool, "remaining_weight", None)
        empty_spool_weight = self._get_empty_spool_weight()

        if remaining_weight is not None:
            measured_weight = (
                remaining_weight + empty_spool_weight
                if empty_spool_weight is not None
                else remaining_weight
            )
            rows.append({
                "mode": "measured",
                "label": _("Measured weight"),
                "value": measured_weight,
            })
            rows.append({
                "mode": "remaining",
                "label": _("Remaining weight"),
                "value": remaining_weight,
            })

        initial_weight = self._get_initial_weight()
        if initial_weight is not None:
            rows.append({
                "mode": "initial",
                "label": _("Initial weight"),
                "value": initial_weight,
            })
        if empty_spool_weight is not None:
            rows.append({
                "mode": "empty",
                "label": _("Empty weight"),
                "value": empty_spool_weight,
            })
        return rows

    def _format_measurement_row_markup(self, row):
        formatted_value = f'{round(row["value"], 2)} g'
        return f'{row["label"]}: {formatted_value}'

    def _get_initial_weight(self):
        return getattr(self.spool, "initial_weight", None)

    def _get_empty_spool_weight(self):
        spool_weight = getattr(self.spool, "spool_weight", None)
        vendor_spool_weight = getattr(getattr(self.spool.filament, "vendor", None), "empty_spool_weight", None)
        return spool_weight if spool_weight is not None else vendor_spool_weight

    def _get_full_filament_weight(self):
        # initial_weight should be net weight
        initial_weight = getattr(self.spool, "initial_weight", None)
        filament_weight = getattr(self.spool.filament, "weight", None)
        if initial_weight is not None:
            return initial_weight
        return filament_weight

    def _get_display_name(self):
        parts = []
        vendor_name = getattr(getattr(self.spool.filament, "vendor", None), "name", None)
        filament_name = getattr(self.spool.filament, "name", None) or getattr(self.spool, "name", None)
        if vendor_name:
            parts.append(vendor_name)
        if filament_name:
            parts.append(filament_name)
        return " - ".join(parts) if parts else self.spool.name

    def _format_last_used(self, value):
        if isinstance(value, datetime):
            if self._config.get_main_config().getboolean("24htime", True):
                return value.astimezone().strftime("%Y-%m-%d %H:%M")
            return value.astimezone().strftime("%Y-%m-%d %I:%M %p")
        return str(value)

    def _build_keypad_panel(self):
        if "keypad" not in self.labels:
            self.labels["keypad"] = Keypad(
                self._screen,
                ok_cb=self._submit_entry,
                cancel_cb=self._hide_numpad,
                entry_max=8,
                error_msg=_("Invalid weight")
            )
            self.labels["keypad"].add_extra_button(
                self._restore_value,
                icon="refresh",
                label=_("Restore")
            )
        return self.labels["keypad"]

    def _build_progress_bar(self):
        used = getattr(self.spool, "used_weight", 0) or 0
        initial = self._get_full_filament_weight() or 1000
        fraction = min(max(used / initial, 0.0), 1.0)
        progress = Gtk.ProgressBar()
        progress.set_fraction(fraction)
        progress.set_text(f"{round(used, 1)} / {round(initial, 1)} g")
        progress.set_show_text(True)
        progress.set_margin_start(10)
        progress.set_margin_end(10)
        progress.set_margin_top(5)
        progress.set_margin_bottom(5)
        progress.set_halign(Gtk.Align.FILL)
        self._progress_bar = progress
        return progress

    def _build_second_panel(self):
        if self.numpad_visible:
            return self._build_keypad_panel()
        return self._build_info_panel()

    def _build_info_panel(self):
        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=10, hexpand=True, vexpand=True)
        box.set_halign(Gtk.Align.FILL)
        box.set_valign(Gtk.Align.FILL)

        # Progress Bar
        progress = self._build_progress_bar()
        box.pack_start(progress, False, False, 0)

        # Info Labels
        info_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=5, hexpand=True)

        vendor = getattr(getattr(getattr(self.spool, "filament", None), "vendor", None), "name", None) or "-"
        material = getattr(getattr(self.spool, "filament", None), "material", None) or "-"
        location = getattr(self.spool, "location", None) or "-"
        last_used = self._format_last_used(getattr(self.spool, "last_used", None))
        archived = _("Yes") if getattr(self.spool, "archived", None) else _("No")
        price = getattr(self.spool, "price", None) or "-"
        info_items = [
            (_("Vendor"), vendor),
            (_("Material"), material),
            (_("Used"), last_used),
            (_("Location"), location),
            (_("Archived"), archived),
            (_("Price"), price),
        ]
        for label_text, value in info_items:
            row = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, hexpand=True)
            lbl = Gtk.Label(label=f"{label_text}: ", halign=Gtk.Align.START)
            val = Gtk.Label(label=str(value), halign=Gtk.Align.START)
            row.pack_start(lbl, False, False, 0)
            row.pack_start(val, True, True, 0)
            info_box.pack_start(row, False, False, 0)

        comment = _("Comment") + ": " + (getattr(self.spool, "comment", None) or "-")
        comment_lbl = Gtk.Label(label=comment, wrap=True, wrap_mode=Pango.WrapMode.WORD_CHAR)
        comment_lbl.set_alignment(0.0, 0.0)
        comment_area = self._gtk.ScrolledWindow()
        comment_area.add(comment_lbl)
        info_box.pack_start(comment_area, True, True, 0)

        box.pack_start(info_box, True, True, 0)
        return box

    @staticmethod
    def _format_entry_weight(value):
        if value is None:
            return ""
        if float(value).is_integer():
            return str(int(value))
        return f"{value:.2f}".rstrip("0").rstrip(".")

    def _submit_entry(self, value):
        if self.spool is None:
            return
        result = {}
        if self.selected_weight_mode == "measured":
            empty_spool_weight = self._get_empty_spool_weight() or 0
            if empty_spool_weight:
                value -= empty_spool_weight
            if value < 0:
                self._screen.show_popup_message(_("Measured weight is below empty spool weight"))
                return
            result = self._screen.spoolman_api.update_spool(
                spool_id=self.spool.id,
                payload={"remaining_weight": value}

            )
        if self.selected_weight_mode == "remaining":
            if value < 0:
                self._screen.show_popup_message(_("Value must be positive"))
                return
            result = self._screen.spoolman_api.update_spool(
                spool_id=self.spool.id,
                payload={"remaining_weight": value}
            )
        if self.selected_weight_mode == "empty":
            if value < 0:
                self._screen.show_popup_message(_("Value must be positive"))
                return
            result = self._screen.spoolman_api.update_spool(
                spool_id=self.spool.id,
                payload={"spool_weight": value}
            )
        if self.selected_weight_mode == "initial":
            if value < 0:
                self._screen.show_popup_message(_("Value must be positive"))
                return
            result = self._screen.spoolman_api.update_spool(
                spool_id=self.spool.id,
                payload={"initial_weight": value}
            )
        if not result:
            self._screen.show_popup_message(_("Error updating filament weight"))
            return

        if result.get("error"):
            self._screen.show_popup_message(
                _("Error updating filament weight") + f"\n{result['error'].get('message', '')}"
            )
            return

        self.spool.remaining_weight = result.get("remaining_weight", value)

        if "used_weight" in result:
            self.spool.used_weight = result["used_weight"]

        if "spool_weight" in result:
            self.spool.spool_weight = result["spool_weight"]

        if "initial_weight" in result:
            self.spool.initial_weight = result["initial_weight"]

        if self._screen.printer is not None and getattr(self._screen.printer, 'active_spool_id', None) == self.spool.id:
            self._screen.printer.active_spool['remaining_weight'] = result.get("remaining_weight")

        self._refresh_values()
        self._screen.show_popup_message(_("Filament weight updated"), 1)
        self._hide_numpad(None)

    def _refresh_values(self):
        for mode, btn in self._measurement_buttons.items():
            for row in self._get_measurement_rows():
                if row["mode"] == mode:
                    btn.set_label(self._format_measurement_row_markup(row))
                    break
        if self._progress_bar is not None:
            used = getattr(self.spool, "used_weight", 0) or 0
            initial = self._get_full_filament_weight() or 1000
            fraction = min(max(used / initial, 0.0), 1.0)
            self._progress_bar.set_fraction(fraction)
            self._progress_bar.set_text(f"{round(used, 1)} / {round(initial, 1)} g")

    def _restore_value(self, entry_val):
        if "keypad" in self.labels:
            self.labels['keypad'].entry.set_text(f"{self.saved_weight:.1f}")

    def back(self):
        return False

    def _show_numpad(self, widget, mode):
        self.selected_weight_mode = mode
        if mode == "measured":
            self.saved_weight = (self.spool.remaining_weight or 0) + (self.spool.spool_weight or 0)
        if mode == "remaining":
            self.saved_weight = self.spool.remaining_weight or 0
        if mode == "empty":
            self.saved_weight = self.spool.spool_weight or 0
        if mode == "initial":
            self.saved_weight = self.spool.initial_weight or 0
        self.numpad_visible = True
        if self._screen.vertical_mode:
            self.grid.remove(self.grid.get_child_at(0, 2))
            self.grid.attach(self._build_second_panel(), 0, 2, 1, 2)
        else:
            self.grid.remove_column(1)
            self.grid.attach(self._build_second_panel(), 1, 0, 1, 1)
        self.grid.show_all()

    def _hide_numpad(self, widget=None):
        self.numpad_visible = False
        if self._screen.vertical_mode:
            numpad = self.grid.get_child_at(0, 2)
            if numpad:
                self.grid.remove(numpad)
            self.grid.attach(self._build_second_panel(), 0, 2, 1, 2)
        else:
            self.grid.remove_column(1)
            self.grid.attach(self._build_second_panel(), 1, 0, 1, 1)
        self.grid.show_all()  # needed?

    def deactivate(self):
        self._hide_numpad()



import logging
import os

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk
from ks_includes.screen_panel import ScreenPanel


class Panel(ScreenPanel):
    def __init__(self, screen, title):
        title = title or _("Shutdown")
        super().__init__(screen, title)

        estop = self._gtk.Button("emergency", _("Emergency Stop"), "color2")
        estop.connect("clicked", self.emergency_stop)

        power_panel = {
            "panel": "power",
        }
        power_shortcut = self._gtk.Button("shutdown", _("Power"), "color2")
        power_shortcut.connect("clicked", self.menu_item_clicked, power_panel)

        poweroff = self._gtk.Button("shutdown", _("Shutdown"), "color1")
        poweroff.connect("clicked", self.reboot_poweroff, "shutdown")

        restart = self._gtk.Button("refresh", _("Restart"), "color3")
        restart.connect("clicked", self.reboot_poweroff, "reboot")

        restart_ks = self._gtk.Button("refresh", _("Restart") + " KlipperScreen", "color3")
        restart_ks.connect("clicked", self._screen.restart_ks)

        lock_screen = self._gtk.Button("lock", _("Lock"), "color3")
        lock_screen.connect("clicked", self._screen.lock_screen.lock)

        self.main = Gtk.Grid(row_homogeneous=True, column_homogeneous=True)
        show_estop = self._printer and self._printer.state not in {'disconnected', 'startup', 'shutdown', 'error'}
        show_power_shortcut = False
        if self.ks_printer_cfg is not None and self._screen._ws.connected:
            power_devices = self.ks_printer_cfg.get("power_devices", "")
            if power_devices and self._printer.get_power_devices():
                show_power_shortcut = True

        if show_estop and show_power_shortcut:
            self.main.attach(estop, 0, 0, 1, 1)
            self.main.attach(power_shortcut, 0, 1, 1, 1)
        elif show_estop:
            self.main.attach(estop, 0, 0, 1, 2)
        elif show_power_shortcut:
            self.main.attach(power_shortcut, 0, 0, 1, 2)
        self.main.attach(restart_ks, 1, 0, 1, 1)
        self.main.attach(lock_screen, 2, 0, 1, 1)
        self.main.attach(poweroff, 1, 1, 1, 1)
        self.main.attach(restart, 2, 1, 1, 1)
        self.content.add(self.main)

    def reboot_poweroff(self, widget, method):
        label = Gtk.Label(wrap=True, hexpand=True, vexpand=True)
        if method == "reboot":
            label.set_label(_("Are you sure you wish to reboot the system?"))
            title = _("Restart")
        else:
            label.set_label(_("Are you sure you wish to shutdown the system?"))
            title = _("Shutdown")
        buttons = []
        if (
            self._screen.apiclient is None
            or "127.0.0.1" in self._screen.apiclient.endpoint
            or "localhost" in self._screen.apiclient.endpoint
        ):
            buttons.append({"name": _("Accept"), "response": Gtk.ResponseType.ACCEPT, "style": 'dialog-primary'})
        else:
            logging.info(self._screen.apiclient.endpoint)
            buttons.extend([
                {"name": _("Host"), "response": Gtk.ResponseType.OK, "style": 'dialog-info'},
                {"name": _("Printer"), "response": Gtk.ResponseType.APPLY, "style": 'dialog-warning'},
                {"name": _("Both"), "response": Gtk.ResponseType.ACCEPT, "style": 'dialog-primary'},
            ])
        buttons.append({"name": _("Cancel"), "response": Gtk.ResponseType.CANCEL, "style": 'dialog-error'})
        self._gtk.Dialog(title, buttons, label, self.reboot_poweroff_confirm, method)

    def reboot_poweroff_confirm(self, dialog, response_id, method):
        self._gtk.remove_dialog(dialog)
        if response_id == Gtk.ResponseType.ACCEPT:
            if method == "reboot":
                self._screen._ws.send_method("machine.reboot")
                os.system("systemctl reboot -i")
            else:
                self._screen._ws.send_method("machine.shutdown")
                os.system("systemctl poweroff -i")
        elif response_id == Gtk.ResponseType.OK:
            if method == "reboot":
                os.system("systemctl reboot -i")
            else:
                os.system("systemctl poweroff -i")
        elif response_id == Gtk.ResponseType.APPLY:
            if method == "reboot":
                self._screen._ws.send_method("machine.reboot")
            else:
                self.turn_off_power_devices()
                self._screen._ws.send_method("machine.shutdown")

    def turn_off_power_devices(self):
        if self.ks_printer_cfg is not None and self._screen._ws.connected:
            power_devices = self.ks_printer_cfg.get("power_devices", "")
            if power_devices and self._printer.get_power_devices():
                logging.info(f"Turning off associated power devices: {power_devices}")
                self._screen.power_devices(widget=None, devices=power_devices, on=False)



# -*- coding: utf-8 -*-
import logging
import os
import pathlib
from functools import lru_cache

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gdk, GdkPixbuf, Gio, Gtk, Pango
from ks_includes.widgets.scroll import CustomScrolledWindow


def find_widget(widget, wanted_type):
    # Returns a widget of wanted_type or None
    if isinstance(widget, wanted_type):
        return widget
    if isinstance(widget, (Gtk.Container, Gtk.Bin, Gtk.Button, Gtk.Alignment, Gtk.Box)):
        for _ in widget.get_children():
            result = find_widget(_, wanted_type)
            if result is not None:
                return result


def format_label(widget, lines=2):
    label = find_widget(widget, Gtk.Label)
    if label is not None:
        label.set_line_wrap_mode(Pango.WrapMode.WORD_CHAR)
        label.set_line_wrap(True)
        label.set_ellipsize(Pango.EllipsizeMode.END)
        label.set_lines(lines)


class KlippyGtk:
    labels = {}

    def __init__(self, screen):
        self.screen = screen
        self.themedir = os.path.join(pathlib.Path(__file__).parent.resolve().parent, "styles", screen.theme, "images")
        self.font_size_type = screen._config.get_main_config().get("font_size", "medium")
        self.width = screen.width
        self.height = screen.height
        self.ultra_tall = (self.height / self.width) >= 3
        self.font_ratio = [28, 42] if self.screen.vertical_mode else [40, 27]
        self.font_size = min(self.width / self.font_ratio[0], self.height / self.font_ratio[1])
        self.img_scale = self.font_size * 2
        self.button_image_scale = 1.38
        self.bsidescale = .65  # Buttons with image at the side
        self.dialog_buttons_height = round(self.height / 5)

        if self.font_size_type == "max":
            self.font_size = self.font_size * 1.06
            self.img_scale = self.img_scale * 0.7
            self.bsidescale = .7
        elif self.font_size_type == "extralarge":
            self.font_size = self.font_size * 1.05
            self.img_scale = self.img_scale * 0.7
            self.bsidescale = 1.0
        elif self.font_size_type == "large":
            self.font_size = self.font_size * 1.025
            self.img_scale = self.img_scale * 0.85
            self.bsidescale = .8
        elif self.font_size_type == "small":
            self.font_size = self.font_size * 0.91
            self.bsidescale = .55
        self.img_width = self.font_size * 3
        self.img_height = self.font_size * 3
        self.titlebar_height = self.font_size * 2
        logging.info(f"Font size: {self.font_size:.1f} ({self.font_size_type})")

        if self.screen.vertical_mode:
            self.action_bar_width = int(self.width)
            self.action_bar_height = int(self.height * .1)
            self.content_width = self.width
            self.content_height = self.height - self.titlebar_height - self.action_bar_height
        else:
            self.action_bar_width = int(self.width * .1)
            self.action_bar_height = int(self.height)
            self.content_width = self.width - self.action_bar_width
            self.content_height = self.height - self.titlebar_height

        self.keyboard_height = self.content_height * 0.5
        if self.ultra_tall:
            self.keyboard_height = self.keyboard_height * 0.5

        self.color_list = {}  # This is set by screen.py init_style()
        for key in self.color_list:
            if "base" in self.color_list[key]:
                rgb = [int(self.color_list[key]['base'][i:i + 2], 16) for i in range(0, 6, 2)]
                self.color_list[key]['rgb'] = rgb

    def get_temp_color(self, device):
        # logging.debug("Color list %s" % self.color_list)
        if device not in self.color_list:
            return False, False

        if 'base' in self.color_list[device]:
            rgb = self.color_list[device]['rgb'].copy()
            if self.color_list[device]['state'] > 0:
                rgb[1] = rgb[1] + self.color_list[device]['hsplit'] * self.color_list[device]['state']
            self.color_list[device]['state'] += 1
            rgb = [x / 255 for x in rgb]
            # logging.debug(f"Assigning color: {device} {rgb}")
        else:
            colors = self.color_list[device]['colors']
            if self.color_list[device]['state'] >= len(colors):
                self.color_list[device]['state'] = 0
            color = colors[self.color_list[device]['state'] % len(colors)]
            rgb = [int(color[i:i + 2], 16) / 255 for i in range(0, 6, 2)]
            self.color_list[device]['state'] += 1
            # logging.debug(f"Assigning color: {device} {rgb} {color}")
        return rgb

    def reset_temp_color(self):
        for key in self.color_list:
            self.color_list[key]['state'] = 0

    def Image(self, image_name=None, width=None, height=None):
        if image_name is None:
            return Gtk.Image()
        pixbuf = self.PixbufFromIcon(image_name, width, height)
        return Gtk.Image.new_from_pixbuf(pixbuf) if pixbuf is not None else Gtk.Image()

    def update_themedir(self, theme):
        self.themedir = os.path.join(pathlib.Path(__file__).parent.resolve().parent, "styles", theme, "images")

    def PixbufFromIcon(self, filename, width=None, height=None):
        width = width if width is not None else self.img_width
        height = height if height is not None else self.img_height
        return self._PixbufFromIcon(filename, self.themedir, int(width), int(height))

    @staticmethod
    @lru_cache(maxsize=500)
    def _PixbufFromIcon(filename, themedir, width, height):
        filename = os.path.join(themedir, filename)
        for ext in ["svg", "png"]:
            file = f"{filename}.{ext}"
            pixbuf = KlippyGtk.PixbufFromFile(file, width, height) if os.path.exists(file) else None
            if pixbuf is not None:
                return pixbuf
        return None

    @staticmethod
    @lru_cache(maxsize=500)
    def PixbufFromFile(filename, width=-1, height=-1):
        try:
            return GdkPixbuf.Pixbuf.new_from_file_at_size(filename, int(width), int(height))
        except Exception as e:
            logging.exception(e)
            logging.error(f"Unable to find image {filename}")
            return None

    def PixbufFromHttp(self, resource, width=-1, height=-1):
        response = self.screen.apiclient.get_thumbnail_stream(resource)
        if response is False:
            return None
        stream = Gio.MemoryInputStream.new_from_data(response, None)
        try:
            pixbuf = GdkPixbuf.Pixbuf.new_from_stream_at_scale(stream, int(width), int(height), True)
        except Exception as e:
            logging.exception(e)
            return None
        stream.close_async(2)
        return pixbuf

    def Button(self, image_name=None, label=None, style=None, scale=None, position=Gtk.PositionType.TOP, lines=2):
        if self.font_size_type == "max" and label is not None:
            image_name = None
        b = Gtk.Button(hexpand=True, vexpand=True, image_position=position, always_show_image=True)
        focusable = self.screen._config.get_main_config().getboolean("keyboard_navigation", fallback=False)
        b.set_can_focus(focusable)
        if label is not None:
            b.set_label(label.replace("\n", " "))
        if image_name is not None:
            b.set_name(image_name)
            if scale is None:
                scale = self.button_image_scale
            if label is None:
                scale = scale * 1.4
            width = height = self.img_scale * scale
            b.set_image(self.Image(image_name, width, height))
            spinner = Gtk.Spinner(width_request=width, height_request=height, no_show_all=True)
            spinner.hide()
            box = find_widget(b, Gtk.Box)
            if box:
                box.add(spinner)

        if label is not None:
            format_label(b, lines)
        if style is not None:
            b.get_style_context().add_class(style)
        b.connect("clicked", self.screen.screensaver.reset_timeout)
        b.connect("clicked", self.screen.lock_screen.reset_timeout)
        return b

    @staticmethod
    def Button_busy(widget, busy):
        spinner = find_widget(widget, Gtk.Spinner)
        image = find_widget(widget, Gtk.Image)
        if busy:
            widget.set_sensitive(False)
            if image:
                widget.set_always_show_image(False)
                image.hide()
            if spinner:
                spinner.start()
                spinner.show()
        else:
            if image:
                widget.set_always_show_image(True)
                image.show()
            if spinner:
                spinner.stop()
                spinner.hide()
            widget.set_sensitive(True)

    def dialog_content_decouple(self, widget, event, dialog):
        self.remove_dialog(dialog)

    def Dialog(self, title, buttons, content, callback=None, *args):
        dialog = Gtk.Dialog(title=title, modal=True, transient_for=self.screen,
                            default_width=self.width, default_height=self.height)
        dialog.set_size_request(self.width, self.height)
        if not self.screen.windowed:
            dialog.fullscreen()

        if buttons:
            max_buttons = 4
            if len(buttons) > max_buttons:
                buttons = buttons[:max_buttons]
            if len(buttons) > 2:
                dialog.get_action_area().set_layout(Gtk.ButtonBoxStyle.EXPAND)
                button_hsize = -1
            else:
                button_hsize = int((self.width / 3))
            for button in buttons:
                style = button['style'] if 'style' in button else 'dialog-default'
                dialog.add_button(button['name'], button['response'])
                button = dialog.get_widget_for_response(button['response'])
                button.set_size_request(button_hsize, self.dialog_buttons_height)
                button.get_style_context().add_class(style)
                format_label(button, 2)
        else:
            # No buttons means clicking anywhere closes the dialog
            content.add_events(Gdk.EventMask.BUTTON_RELEASE_MASK)
            content.connect("button-release-event", self.dialog_content_decouple, dialog)
            dialog.add_events(Gdk.EventMask.BUTTON_RELEASE_MASK)
            dialog.connect("button-release-event", self.remove_dialog)

        dialog.connect("response", self.screen.screensaver.reset_timeout)
        dialog.connect("response", self.screen.lock_screen.reset_timeout)
        dialog.connect("response", callback, *args)
        dialog.get_style_context().add_class("dialog")

        content_area = dialog.get_content_area()
        content_area.set_margin_start(10)
        content_area.set_margin_end(5)
        content_area.set_margin_top(5)
        content_area.set_margin_bottom(0)
        content_area.add(content)

        dialog.show_all()
        # Change cursor to blank
        self.set_cursor(show=self.screen.show_cursor, window=dialog.get_window())

        self.screen.dialogs.append(dialog)
        logging.info(f"Showing dialog {dialog.get_title()} {dialog.get_size()}")
        return dialog

    def remove_dialog(self, dialog, *args):
        if not isinstance(dialog, Gtk.Dialog):
            logging.error(f"Invalid dialog: {dialog}")
            return
        if self.screen.updating:
            return
        if dialog == self.screen.confirm:
            self.screen.confirm = None
        dialog.destroy()
        if dialog in self.screen.dialogs:
            logging.info("Removing Dialog")
            self.screen.dialogs.remove(dialog)
            return
        logging.debug(f"Cannot remove dialog {dialog}")

    def ScrolledWindow(self, steppers=True, **kwargs):
        steppers = steppers and self.screen._config.get_main_config().getboolean("show_scroll_steppers", fallback=False)
        return CustomScrolledWindow(steppers, **kwargs)

    def set_cursor(self, show: bool, window: Gdk.Window):
        if show:
            window.set_cursor(
                Gdk.Cursor.new_for_display(Gdk.Display.get_default(), Gdk.CursorType.ARROW))
            if not self.screen.wayland:
                os.system("xsetroot  -cursor_name  arrow")
        else:
            window.set_cursor(
                Gdk.Cursor.new_for_display(Gdk.Display.get_default(), Gdk.CursorType.BLANK_CURSOR))
            if not self.screen.wayland:
                os.system("xsetroot  -cursor ks_includes/emptyCursor.xbm ks_includes/emptyCursor.xbm")



import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk


class PrintListItem(Gtk.FlowBoxChild):
    def __init__(self):
        super().__init__()
        self.date = None
        self.size = None
        self.dir = 0
        self.path = None

    def set_date(self, date):
        self.date = date

    def set_size(self, size):
        self.size = size

    def set_as_dir(self, is_dir: bool):
        self.dir = -1 if is_dir else 0

    def set_path(self, path):
        self.path = path

    def get_date(self):
        return self.date

    def get_size(self):
        return self.size

    def get_is_dir(self):
        return self.dir

    def get_path(self):
        return self.path



import logging
import math
import re

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk
from ks_includes.KlippyGcodes import KlippyGcodes
from ks_includes.screen_panel import ScreenPanel


# Find the screw closest to the point,
# but return None if the distance is above max_distance.
# If remove is set to true, the screw is also removed
# from the list of passed in screws.
def find_closest(screws, point, max_distance):
    if len(screws) == 0:
        return None
    closest = screws[0]
    min_distance = math.hypot(closest[0] - point[0], closest[1] - point[1])
    for screw in screws[1:]:
        distance = math.hypot(screw[0] - point[0], screw[1] - point[1])
        if distance < min_distance:
            closest = screw
            min_distance = distance

    if min_distance > max_distance:
        return None
    screws.remove(closest)
    return closest


class Panel(ScreenPanel):

    def __init__(self, screen, title):
        title = title or _("Bed Level")
        super().__init__(screen, title)
        self.screw_positions = {}
        self.screws = []
        self.x_offset = 0
        self.y_offset = 0
        self.buttons = {'dm': self._gtk.Button("motor-off", _("Disable Motors"), "color3")}
        self.buttons['dm'].connect("clicked", self.disable_motors)
        rotation = 0
        self.probe_z_height = 0
        self.lift_speed = 5
        self.horizontal_move_z = 5
        self.horizontal_speed = 50
        invert_x = invert_y = False

        grid = Gtk.Grid(row_homogeneous=True, column_homogeneous=True)
        grid.attach(self.buttons['dm'], 0, 0, 1, 1)

        if "screws_tilt_adjust" in self._printer.get_config_section_list():
            self.buttons['screws'] = self._gtk.Button("refresh", _("Screws Adjust"), "color4")
            self.buttons['screws'].connect("clicked", self.screws_tilt_calculate)
            grid.attach(self.buttons['screws'], 0, 1, 1, 1)

            self.screws = self._get_screws("screws_tilt_adjust")
            logging.info(f"screws_tilt_adjust: {self.screws}")

            probe = self._printer.get_probe()
            if probe:
                if "x_offset" in probe:
                    self.x_offset = float(probe['x_offset'])
                if "y_offset" in probe:
                    self.y_offset = float(probe['y_offset'])
                logging.debug(f"offset X: {self.x_offset} Y: {self.y_offset}")
            # bed_screws uses NOZZLE positions
            # screws_tilt_adjust uses PROBE positions and
            # to be offseted for the buttons to work equal to bed_screws
            new_screws = [
                [screw[0] + self.x_offset, screw[1] + self.y_offset]
                for screw in self.screws
            ]

            self.screws = new_screws
            logging.info(f"screws with offset: {self.screws}")
        elif "bed_screws" in self._printer.get_config_section_list():
            self.screws = self._get_screws("bed_screws")
            logging.info(f"bed_screws: {self.screws}")

        # KS config
        if self.ks_printer_cfg is not None:
            rotation = self.ks_printer_cfg.getint("screw_rotation", 0)
            if rotation not in (0, 90, 180, 270):
                self._screen.show_popup_message(_("Rotation invalid") + f" {rotation} \n")
                logging.info(f"Rotation invalid: {rotation}")
                rotation = 0
            logging.info(f"Rotation: {rotation}")
            invert_x = self._config.get_config()['main'].getboolean("invert_x", False)
            invert_y = self._config.get_config()['main'].getboolean("invert_y", False)
            logging.info(f"Inversion X: {invert_x} Y: {invert_y}")

        # get dimensions
        x_positions = {x[0] for x in self.screws}
        y_positions = {y[1] for y in self.screws}
        logging.info(f"X: {x_positions}\nY: {y_positions}")

        min_x = min(x_positions)
        max_x = max(x_positions)
        mid_x = round((min_x + max_x) / 2)

        min_y = min(y_positions)
        max_y = max(y_positions)
        mid_y = round((min_y + max_y) / 2)

        max_distance = math.floor(min(max_x - min_x, max_y - min_y) / 3)

        logging.debug(f"Using max_distance: {max_distance} to fit: {len(self.screws)} screws.")

        remaining_screws = self.screws[:]

        # The order here it's important because the rotation function will
        # shift the values according to the angle of rotation
        self.screw_positions = {
            'bl': find_closest(remaining_screws, (min_x, max_y), max_distance),
            'fm': find_closest(remaining_screws, (mid_x, min_y), max_distance),
            'br': find_closest(remaining_screws, (max_x, max_y), max_distance),
            'lm': find_closest(remaining_screws, (min_x, mid_y), max_distance),
            'fr': find_closest(remaining_screws, (max_x, min_y), max_distance),
            'bm': find_closest(remaining_screws, (mid_x, max_y), max_distance),
            'fl': find_closest(remaining_screws, (min_x, min_y), max_distance),
            'rm': find_closest(remaining_screws, (max_x, mid_y), max_distance),
        }

        if invert_x and invert_y:
            rotation = (rotation + 180) % 360
            invert_x = invert_y = False
        if rotation != 0:
            self.screw_positions = self.map_rotation(self.screw_positions, rotation)
            logging.info(f"Rotated: {rotation}")
        if invert_x or invert_y:
            self.screw_positions = self.map_invert(self.screw_positions, invert_x, invert_y)

        self.screw_positions['center'] = find_closest(remaining_screws, (mid_x, mid_y), max_distance)

        if len(remaining_screws) != 0:
            found = []
            for pos in self.screw_positions:
                if self.screw_positions[pos]:
                    found.append(pos)
            logging.debug(f"Found: {found}")
            logging.debug(f"Screws not used: {remaining_screws}")
            if len(self.screws) > 9:
                error_msg = _("This panel supports up-to 9 screws in a 3x3 Grid")
            else:
                error_msg = _("It's possible that the configuration is not correct")
            self._screen.show_popup_message(_("Screws not used:") + f" {remaining_screws} \n" +
                                            error_msg, 2)

        logging.debug(f"Using {len(self.screws) - len(remaining_screws)}/{len(self.screws)}-screw locations")

        button_scale = 2

        self.buttons['bl'] = self._gtk.Button("bed-level-t-l", scale=button_scale)
        self.buttons['br'] = self._gtk.Button("bed-level-t-r", scale=button_scale)
        self.buttons['fl'] = self._gtk.Button("bed-level-b-l", scale=button_scale)
        self.buttons['fr'] = self._gtk.Button("bed-level-b-r", scale=button_scale)
        self.buttons['lm'] = self._gtk.Button("bed-level-l-m", scale=button_scale)
        self.buttons['rm'] = self._gtk.Button("bed-level-r-m", scale=button_scale)
        self.buttons['fm'] = self._gtk.Button("bed-level-b-m", scale=button_scale)
        self.buttons['bm'] = self._gtk.Button("bed-level-t-m", scale=button_scale)
        self.buttons['center'] = self._gtk.Button("bed-level-center", scale=button_scale)

        screw_layout_map = {
            'fr': [3, 2, 1, 1],
            'fm': [2, 2, 1, 1],
            'fl': [1, 2, 1, 1],
            'rm': [3, 1, 1, 1],
            'br': [3, 0, 1, 1],
            'bm': [2, 0, 1, 1],
            'bl': [1, 0, 1, 1],
            'lm': [1, 1, 1, 1],
            'center': [2, 1, 1, 1],
        }

        bedgrid = Gtk.Grid()
        for pos in screw_layout_map:
            bedgrid.attach(self.buttons[pos], *screw_layout_map[pos])
            self.buttons[pos].set_no_show_all(True)
            if pos in self.screw_positions and self.screw_positions[pos]:
                self.buttons[pos].show()

        for layout_pos in self.screw_positions:
            self.buttons[layout_pos].connect("clicked", self.go_to_position, self.screw_positions[layout_pos])

        remove_list = []
        for screw in self.screw_positions:
            if self.screw_positions[screw] is None:
                remove_list.append(screw)
        for screw in remove_list:
            self.screw_positions.pop(screw)

        logging.info(f"screw_positions: {self.screw_positions}")

        grid.attach(bedgrid, 1, 0, 3, 2)
        self.content.add(grid)

    @staticmethod
    def map_invert(positions, invert_x, invert_y):
        if invert_x:
            return {
                'fr': positions['fl'],
                'fm': positions['fm'],
                'fl': positions['fr'],
                'rm': positions['lm'],
                'bl': positions['br'],
                'bm': positions['bm'],
                'br': positions['bl'],
                'lm': positions['rm']
            }
        if invert_y:
            return {
                'fr': positions['br'],
                'fm': positions['bm'],
                'fl': positions['bl'],
                'rm': positions['rm'],
                'bl': positions['fl'],
                'bm': positions['fm'],
                'br': positions['fr'],
                'lm': positions['lm']
            }
        return positions

    @staticmethod
    def map_rotation(positions, angle):
        angle %= 360
        shift = (angle // 90) * 2
        rotated_positions = {}
        keys = list(positions.keys())
        for i, key in enumerate(keys):
            new_key = keys[(i + shift) % len(keys)]
            rotated_positions[new_key] = positions[key]
        return rotated_positions

    def home(self):
        # Test if all axes have been homed. Home if necessary.
        if self._printer.get_stat("toolhead", "homed_axes") != "xyz":
            self._screen._ws.klippy.gcode_script("G28")
            # do Z_TILT_CALIBRATE if applicable.
            if self._printer.config_section_exists("z_tilt"):
                self._screen._ws.klippy.gcode_script("Z_TILT_ADJUST")

    def go_to_position(self, widget, position):
        self.home()
        logging.debug(f"Going to position: {position}")
        script = [
            f"{KlippyGcodes.MOVE_ABSOLUTE}",
            f"G1 Z{self.horizontal_move_z} F{self.lift_speed * 60}\n",
            f"G1 X{position[0]} Y{position[1]} F{self.horizontal_speed * 60}\n",
            f"G1 Z{self.probe_z_height} F{self.lift_speed * 60}\n"
        ]
        self._screen._send_action(widget, "printer.gcode.script", {"script": "\n".join(script)})

    def disable_motors(self, widget):
        self._screen._send_action(widget, "printer.gcode.script", {"script": "M18"})

    def process_busy(self, busy):
        for button in self.buttons:
            if button == "screws":
                continue
            self.buttons[button].set_sensitive((not busy))

    def process_update(self, action, data):
        if 'idle_timeout' in data:
            self.process_busy(data['idle_timeout']['state'].lower() == "printing")
        if action != "notify_status_update":
            return
        if "screws_tilt_adjust" in data:
            if "error" in data["screws_tilt_adjust"]:
                self.buttons['screws'].set_sensitive(True)
                logging.info("Error reported by screws_tilt_adjust")
            if "results" in data["screws_tilt_adjust"]:
                section = self._printer.get_config_section('screws_tilt_adjust')
                for screw, result in data["screws_tilt_adjust"]["results"].items():
                    logging.info(f"{screw} {result['sign']} {result['adjust']}")
                    if screw not in section:
                        logging.error(f"{screw} not found in {section}")
                        continue
                    x, y = section[screw].split(',')
                    x = float(x) + self.x_offset
                    y = float(y) + self.y_offset
                    for key, value in self.screw_positions.items():
                        if value and x == value[0] and y == value[1]:
                            logging.debug(f"X: {x} Y: {y} Adjust: {result['adjust']} Pos: {key}")
                            if result['is_base']:
                                logging.info(f"{screw} is the Reference")
                                self.buttons[key].set_label(_("Reference"))
                            else:
                                self.buttons[key].set_label(f"{result['sign']} {result['adjust']}")
                            if int(result['adjust'].split(':')[0]) == 0 and int(result['adjust'].split(':')[1]) < 6:
                                self.buttons[key].set_image(self._gtk.Image('complete'))
                            else:
                                self.buttons[key].set_image(self._gtk.Image(result['sign'].lower()))

    def _get_screws(self, config_section_name):
        screws = []
        config_section = self._printer.get_config_section(config_section_name)
        logging.debug(config_section_name)
        for item in config_section:
            logging.debug(f"{item}: {config_section[item]}")
            if item == 'probe_speed':
                self.lift_speed = float(config_section[item])
            elif item == 'speed':
                self.horizontal_speed = float(config_section[item])
            elif item == 'horizontal_move_z':
                self.horizontal_move_z = float(config_section[item])
            elif item == 'probe_height':
                self.probe_z_height = float(config_section[item])
            else:
                result = re.match(r"([\-0-9\.]+)\s*,\s*([\-0-9\.]+)", config_section[item])
                if result:
                    screws.append([
                        float(result[1]),
                        float(result[2])
                    ])
        return sorted(screws, key=lambda s: (float(s[1]), float(s[0])))

    def screws_tilt_calculate(self, widget):
        self.home()
        self._screen._send_action(widget, "printer.gcode.script", {"script": "SCREWS_TILT_CALCULATE"})



import logging
import re

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk
from ks_includes.KlippyGcodes import KlippyGcodes
from ks_includes.screen_panel import ScreenPanel


class Panel(ScreenPanel):
    distances = [".1", ".5", "1", "5", "10", "25", "50"]
    distance = distances[-2]

    def __init__(self, screen, title):
        title = title or _("Move")
        super().__init__(screen, title)

        if self.ks_printer_cfg is not None:
            dis = self.ks_printer_cfg.get("move_distances", "")
            if re.match(r"^[0-9,\.\s]+$", dis):
                dis = [str(i.strip()) for i in dis.split(",")]
                if 1 < len(dis) <= 7:
                    self.distances = dis
                    self.distance = self.distances[-2]

        self.settings = {}
        self.menu.append("move_menu")
        self.buttons = {
            "x+": self._gtk.Button("arrow-right", "X+", "color1"),
            "x-": self._gtk.Button("arrow-left", "X-", "color1"),
            "y+": self._gtk.Button("arrow-up", "Y+", "color2"),
            "y-": self._gtk.Button("arrow-down", "Y-", "color2"),
            "z+": self._gtk.Button("z-farther", "Z+", "color3"),
            "z-": self._gtk.Button("z-closer", "Z-", "color3"),
            "home": self._gtk.Button("home", _("Home"), "color4"),
            "motors_off": self._gtk.Button("motor-off", _("Disable Motors"), "color4"),
        }
        self.buttons["x+"].connect("clicked", self.move, "X", "+")
        self.buttons["x-"].connect("clicked", self.move, "X", "-")
        self.buttons["y+"].connect("clicked", self.move, "Y", "+")
        self.buttons["y-"].connect("clicked", self.move, "Y", "-")
        self.buttons["z+"].connect("clicked", self.move, "Z", "+")
        self.buttons["z-"].connect("clicked", self.move, "Z", "-")
        self.buttons["home"].connect("clicked", self.home)
        script = {"script": "M18"}
        self.buttons["motors_off"].connect(
            "clicked",
            self._screen._confirm_send_action,
            _("Are you sure you wish to disable motors?"),
            "printer.gcode.script",
            script,
        )
        adjust = self._gtk.Button(
            "settings", None, "color2", 1, Gtk.PositionType.LEFT, 1
        )
        adjust.connect("clicked", self.load_menu, "options", _("Settings"))
        adjust.set_hexpand(False)

        xm = "x-"
        xp = "x+"
        ym = "y-"
        yp = "y+"
        zm = "z-"
        zp = "z+"

        rotation = (
            0
            if self.ks_printer_cfg is None
            else self.ks_printer_cfg.getint("screw_rotation", 0)
        )
        if rotation == 90:
            xm, yp, xp, ym = ym, xm, yp, xp
        elif rotation == 180:
            xm, yp, xp, ym = xp, ym, xm, yp
        elif rotation == 270:
            xm, yp, xp, ym = yp, xp, ym, xm

        if self._config.get_config()["main"].getboolean("invert_z", False):
            zm, zp = zp, zm

        scale = self._gtk.img_scale * self._gtk.button_image_scale
        self.buttons[xm].set_image(self._gtk.Image("arrow-left", scale, scale))
        self.buttons[yp].set_image(self._gtk.Image("arrow-up", scale, scale))
        self.buttons[xp].set_image(self._gtk.Image("arrow-right", scale, scale))
        self.buttons[ym].set_image(self._gtk.Image("arrow-down", scale, scale))

        grid = Gtk.Grid(row_homogeneous=True, column_homogeneous=True)
        if self._screen.vertical_mode:
            grid.attach(self.buttons[xm], 0, 1, 1, 1)
            grid.attach(self.buttons[xp], 2, 1, 1, 1)
            grid.attach(self.buttons[ym], 1, 1, 1, 1)
            grid.attach(self.buttons[yp], 1, 0, 1, 1)
            grid.attach(self.buttons[zm], 0, 2, 1, 1)
            grid.attach(self.buttons[zp], 2, 2, 1, 1)
            grid.attach(adjust, 1, 2, 1, 1)
        else:
            grid.attach(self.buttons[xm], 0, 1, 1, 1)
            grid.attach(self.buttons[xp], 2, 1, 1, 1)
            grid.attach(self.buttons[ym], 1, 1, 1, 1)
            grid.attach(self.buttons[yp], 1, 0, 1, 1)
            grid.attach(self.buttons[zm], 3, 1, 1, 1)
            grid.attach(self.buttons[zp], 3, 0, 1, 1)

        grid.attach(self.buttons["home"], 0, 0, 1, 1)
        grid.attach(self.buttons["motors_off"], 2, 0, 1, 1)

        distgrid = Gtk.Grid()
        for j, i in enumerate(self.distances):
            self.labels[i] = self._gtk.Button(label=i)
            self.labels[i].set_direction(Gtk.TextDirection.LTR)
            self.labels[i].connect("clicked", self.change_distance, i)
            ctx = self.labels[i].get_style_context()
            ctx.add_class("horizontal_togglebuttons")
            if i == self.distance:
                ctx.add_class("horizontal_togglebuttons_active")
            distgrid.attach(self.labels[i], j, 0, 1, 1)

        for p in ("pos_x", "pos_y", "pos_z"):
            self.labels[p] = Gtk.Label()
        self.labels["move_dist"] = Gtk.Label(label=_("Move Distance (mm)"))

        bottomgrid = Gtk.Grid(row_homogeneous=True, column_homogeneous=True)
        bottomgrid.set_direction(Gtk.TextDirection.LTR)
        bottomgrid.attach(self.labels["pos_x"], 0, 0, 1, 1)
        bottomgrid.attach(self.labels["pos_y"], 1, 0, 1, 1)
        bottomgrid.attach(self.labels["pos_z"], 2, 0, 1, 1)
        bottomgrid.attach(self.labels["move_dist"], 0, 1, 3, 1)
        if not self._screen.vertical_mode:
            bottomgrid.attach(adjust, 3, 0, 1, 2)

        self.labels["move_menu"] = Gtk.Grid(
            row_homogeneous=True, column_homogeneous=True
        )
        self.labels["move_menu"].attach(grid, 0, 0, 1, 3)
        self.labels["move_menu"].attach(bottomgrid, 0, 3, 1, 1)
        self.labels["move_menu"].attach(distgrid, 0, 4, 1, 1)

        self.content.add(self.labels["move_menu"])

        printer_cfg = self._printer.get_config_section("printer")
        # The max_velocity parameter is not optional in klipper config.
        # The minimum is 1, but least 2 values are needed to create a scale
        max_velocity = max(int(float(printer_cfg["max_velocity"])), 2)
        if "max_z_velocity" in printer_cfg:
            self.max_z_velocity = max(int(float(printer_cfg["max_z_velocity"])), 2)
        else:
            self.max_z_velocity = max_velocity

        configurable_options = [
            {
                "invert_x": {
                    "section": "main",
                    "name": _("Invert X"),
                    "type": "binary",
                    "tooltip": _("This will affect screw positions and mesh graph"),
                    "value": "False",
                    "callback": self.reinit_panels,
                }
            },
            {
                "invert_y": {
                    "section": "main",
                    "name": _("Invert Y"),
                    "type": "binary",
                    "tooltip": _("This will affect screw positions and mesh graph"),
                    "value": "False",
                    "callback": self.reinit_panels,
                }
            },
            {
                "invert_z": {
                    "section": "main",
                    "name": _("Invert Z"),
                    "tooltip": _(
                        "Swaps buttons if they are on top of each other, affects other panels"
                    ),
                    "type": "binary",
                    "value": "False",
                    "callback": self.reinit_move,
                }
            },
            {
                "move_speed_xy": {
                    "section": "main",
                    "name": _("XY Speed (mm/s)"),
                    "type": "scale",
                    "tooltip": _("Only for the move panel"),
                    "value": "50",
                    "range": [1, max_velocity],
                    "step": 1,
                }
            },
            {
                "move_speed_z": {
                    "section": "main",
                    "name": _("Z Speed (mm/s)"),
                    "type": "scale",
                    "tooltip": _("Only for the move panel"),
                    "value": "10",
                    "range": [1, self.max_z_velocity],
                    "step": 1,
                }
            },
        ]

        self.labels["options_menu"] = self._gtk.ScrolledWindow()
        self.labels["options"] = Gtk.Grid()
        self.labels["options_menu"].add(self.labels["options"])
        self.options = {}
        for option in configurable_options:
            name = list(option)[0]
            self.options.update(
                self.add_option("options", self.settings, name, option[name])
            )

    def reinit_panels(self, value):
        self._screen.panels_reinit.append("bed_level")
        self._screen.panels_reinit.append("bed_mesh")

    def reinit_move(self, widget):
        self._screen.panels_reinit.append("move")
        self._screen.panels_reinit.append("zcalibrate")
        self.menu.clear()

    def process_update(self, action, data):
        if action != "notify_status_update":
            return
        if "toolhead" in data and "max_velocity" in data["toolhead"]:
            max_vel = max(int(float(data["toolhead"]["max_velocity"])), 2)
            adj = self.options["move_speed_xy"].get_adjustment()
            adj.set_upper(max_vel)
        if (
            "gcode_move" in data
            or "toolhead" in data
            and "homed_axes" in data["toolhead"]
        ):
            homed_axes = self._printer.get_stat("toolhead", "homed_axes")
            for i, axis in enumerate(("x", "y", "z")):
                if axis not in homed_axes:
                    self.labels[f"pos_{axis}"].set_text(f"{axis.upper()}: ?")
                elif "gcode_move" in data and "gcode_position" in data["gcode_move"]:
                    self.labels[f"pos_{axis}"].set_text(
                        f"{axis.upper()}: {data['gcode_move']['gcode_position'][i]:.2f}"
                    )

    def change_distance(self, widget, distance):
        logging.info(f"### Distance {distance}")
        self.labels[f"{self.distance}"].get_style_context().remove_class(
            "horizontal_togglebuttons_active"
        )
        self.labels[f"{distance}"].get_style_context().add_class(
            "horizontal_togglebuttons_active"
        )
        self.distance = distance

    def move(self, widget, axis, direction):
        axis = axis.lower()
        if (
            self._config.get_config()["main"].getboolean(f"invert_{axis}", False)
            and axis != "z"
        ):
            direction = "-" if direction == "+" else "+"

        dist = f"{direction}{self.distance}"
        config_key = "move_speed_z" if axis == "z" else "move_speed_xy"
        speed = (
            None
            if self.ks_printer_cfg is None
            else self.ks_printer_cfg.getint(config_key, None)
        )
        if speed is None:
            speed = self._config.get_config()["main"].getint(config_key, self.max_z_velocity)
        speed = 60 * max(1, speed)
        script = f"{KlippyGcodes.MOVE_RELATIVE}\nG0 {axis}{dist} F{speed}"
        self._screen._send_action(widget, "printer.gcode.script", {"script": script})
        if self._printer.get_stat("gcode_move", "absolute_coordinates"):
            self._screen._ws.klippy.gcode_script("G90")

    def home(self, widget):
        if "delta" in self._printer.get_config_section("printer")["kinematics"]:
            self._screen._send_action(widget, "printer.gcode.script", {"script": "G28"})
            return
        name = "homing"
        disname = self._screen._config.get_menu_name("move", name)
        menuitems = self._screen._config.get_menu_items("move", name)
        self._screen.show_panel("menu", disname, items=menuitems)



import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, Gdk


class CustomScrolledWindow(Gtk.ScrolledWindow):
    def __init__(self, steppers=False, **kwargs):
        args = {
            "hexpand": True,
            "vexpand": True,
            "overlay_scrolling": False
        }
        args.update(kwargs)
        super().__init__(**args)
        self.add_events(Gdk.EventMask.BUTTON_PRESS_MASK |
                        Gdk.EventMask.TOUCH_MASK |
                        Gdk.EventMask.BUTTON_RELEASE_MASK)
        if steppers:
            self.get_vscrollbar().get_style_context().add_class("with-steppers")



import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk


class BedMap(Gtk.DrawingArea):
    def __init__(self, font_size, bm):
        super().__init__()
        self.set_hexpand(True)
        self.set_vexpand(True)
        self.connect('draw', self.draw_graph)
        self.font_size = font_size
        self.font_spacing = round(self.font_size * 1.5)
        self.bm = list(reversed(bm)) if bm is not None else None
        self.invert_x = False
        self.invert_y = False
        self.rotation = 0
        self.mesh_min = [0, 0]
        self.mesh_max = [0, 0]
        self.mesh_radius = 0

    def update_bm(self, bm, radius=None):
        if not bm:
            self.bm = None
            return

        for key, value in bm.items():
            if key == 'profiles':
                continue
        if radius:
            self.mesh_radius = float(radius)
        if 'mesh_min' in bm:
            self.mesh_min = bm['mesh_min']
        elif 'min_x' in bm and 'min_y' in bm:
            self.mesh_min = (float(bm['min_x']), float(bm['min_y']))
        if 'mesh_max' in bm:
            self.mesh_max = bm['mesh_max']
        elif 'max_x' in bm and 'max_y' in bm:
            self.mesh_max = (float(bm['max_x']), float(bm['max_y']))
        if 'probed_matrix' in bm:
            bm = bm['probed_matrix']
        elif 'points' in bm:
            bm = self.transform_points_to_matrix(bm['points'])
        else:
            self.bm = None
            return

        if self.invert_x and self.invert_y:
            self.rotation = (self.rotation + 180) % 360
            self.invert_x = self.invert_y = False
        if self.invert_x:
            new_max = [self.mesh_min[0], self.mesh_max[1]]
            new_min = [self.mesh_max[0], self.mesh_min[1]]
            self.mesh_max = new_max
            self.mesh_min = new_min
            self.bm = [list(reversed(b)) for b in list(reversed(bm))]
        if self.invert_y:
            new_max = [self.mesh_max[0], self.mesh_min[1]]
            new_min = [self.mesh_min[0], self.mesh_max[1]]
            self.mesh_max = new_max
            self.mesh_min = new_min
            self.bm = list(bm)
        else:
            self.bm = list(reversed(bm))

        if self.rotation in (90, 180, 270):
            self.bm = self.rotate_matrix(self.bm)

    @staticmethod
    def transform_points_to_matrix(points):
        rows = points.strip().split('\n')
        return [list(map(float, row.split(','))) for row in rows]

    def rotate_matrix(self, matrix):
        if self.rotation == 90:
            new_max = [self.mesh_max[1], self.mesh_min[0]]
            new_min = [self.mesh_min[1], self.mesh_max[0]]
            self.mesh_max = new_max
            self.mesh_min = new_min
            return [list(row) for row in zip(*matrix[::-1])]
        elif self.rotation == 180:
            new_max = [self.mesh_min[0], self.mesh_min[1]]
            new_min = [self.mesh_max[0], self.mesh_max[1]]
            self.mesh_max = new_max
            self.mesh_min = new_min
            return [list(row)[::-1] for row in matrix[::-1]]
        elif self.rotation == 270:
            new_max = [self.mesh_min[1], self.mesh_max[0]]
            new_min = [self.mesh_max[1], self.mesh_min[0]]

            self.mesh_max = new_max
            self.mesh_min = new_min
            return [list(row) for row in zip(*matrix)][::-1]

    def draw_graph(self, da, ctx):
        width = da.get_allocated_width()
        height = da.get_allocated_height()
        gwidth = int(width - self.font_size * 2.2)
        gheight = int(height - self.font_size * 1.8)
        # Styling
        ctx.set_line_width(1)
        ctx.set_font_size(self.font_size)

        if self.bm is None:
            ctx.move_to(self.font_spacing, height / 2)
            ctx.set_source_rgb(0.5, 0.5, 0.5)
            ctx.show_text(_("No mesh has been loaded"))
            ctx.stroke()
            return

        text_side_top = [0, self.font_size]
        text_side_bottom = [0, height - int(self.font_size * 2)]
        text_side_middle = [self.font_size, (text_side_top[1] + text_side_bottom[1]) / 2]
        text_bottom_left = [self.font_size * 1.8, height - int(self.font_size / 2)]
        text_bottom_right = [width - int(self.font_size * 2.2), height - int(self.font_size / 2)]
        text_bottom_middle = [int(self.font_size * 2 + gwidth / 2), height - int(self.font_size / 2)]

        ctx.set_source_rgb(0.5, 0.5, 0.5)
        ctx.move_to(*text_side_middle)
        ctx.show_text(f"{'Y' if self.rotation in (0, 180) else 'X'}")
        ctx.move_to(*text_bottom_middle)
        ctx.show_text(f"{'X' if self.rotation in (0, 180) else 'Y'}")
        ctx.stroke()

        # min and max axis labels
        ctx.move_to(*text_side_top)
        ctx.show_text(f"{self.mesh_max[1]:.0f}".rjust(4, " "))
        ctx.move_to(*text_side_bottom)
        ctx.show_text(f"{self.mesh_min[1]:.0f}".rjust(4, " "))
        ctx.move_to(*text_bottom_left)
        ctx.show_text(f"{self.mesh_min[0]:.0f}".rjust(4, " "))
        ctx.move_to(*text_bottom_right)
        ctx.show_text(f"{self.mesh_max[0]:.0f}".rjust(4, " "))
        ctx.stroke()

        rows = len(self.bm)
        columns = len(self.bm[0])
        for i, row in enumerate(self.bm):
            ty = (gheight / rows * i)
            by = ty + gheight / rows
            column: float
            for j, column in enumerate(row):
                if self.mesh_radius > 0 and self.round_bed_skip(i, j, row, rows, columns):
                    continue
                lx = (gwidth / columns * j) + self.font_size * 2.2
                rx = lx + gwidth / columns
                # Colors
                ctx.set_source_rgb(*self.colorbar(column))
                ctx.move_to(lx, ty)
                ctx.line_to(lx, by)
                ctx.line_to(rx, by)
                ctx.line_to(rx, ty)
                ctx.close_path()
                ctx.fill()
                ctx.stroke()
                # Numbers
                if gwidth / columns < self.font_size * 3:
                    continue
                ctx.set_source_rgb(0, 0, 0)
                if column > 0:
                    ctx.move_to((lx + rx) / 2 - self.font_size, (ty + by + self.font_size) / 2)
                else:
                    ctx.move_to((lx + rx) / 2 - self.font_size * 1.2, (ty + by + self.font_size) / 2)
                ctx.show_text(f"{column:.2f}")
                ctx.stroke()

    @staticmethod
    def round_bed_skip(i, j, row, rows, columns):
        if columns <= 3:
            return False
        if i != rows // 2 and j != columns // 2:
            # Skip if the value is equal to the next but verify that this also happens on the other side
            if j < columns // 2 and row[j] == row[j + 1] and row[columns - 1] == row[columns - 1 - j]:
                return True
            if j > columns // 2 and row[j] == row[j - 1] and row[0] == row[columns - j]:
                return True
        return False

    @staticmethod
    def colorbar(value: float):
        rmax = 0.25
        color = min(1, max(0, 1 - 1 / rmax * abs(value)))
        if value > 0:
            return [1, color, color]
        if value < 0:
            return [color, color, 1]
        return [1, 1, 1]

    def set_inversion(self, x=False, y=False):
        self.invert_x = x
        self.invert_y = y

    def set_rotation(self, rotation=0):
        self.rotation = rotation



import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk


class AutoGrid(Gtk.Grid):
    """
    A subclass of Gtk.Grid that auto-arranges its children on init

    Args:
        items (list): All the widgets to be arranged.
        max_columns (int: 4): The maximum number of columns, up to 4.
        expand_last (bool: False): expand the last widget to double width.
        vertical (bool: False): optimize for vertical orientation.

    Example:
        grid = Autogrid([Gtk.Button(), Gtk.Button()])
    """

    def __init__(self, items=None, max_columns=None, expand_last=False, vertical=False):
        super().__init__(row_homogeneous=True, column_homogeneous=True)
        if not max_columns:
            max_columns = 3 if vertical else 4
        self.expand_last = expand_last
        if not items:
            return
        length = len(items)
        if vertical and length < 4:
            # Arrange 1 x 4
            columns = 1
        elif length in {4, 2}:
            # Arrange 2 x 2
            columns = min(2, max_columns)
        elif length in {3, 5, 6}:
            # Arrange 3 x 2
            columns = min(3, max_columns)
        else:
            # Arrange 4 x n
            columns = min(4, max_columns)

        for i, widget in enumerate(items):
            col = i % columns
            row = int(i / columns)
            if self.expand_last and (i + 1) == length and length % 2 == 1:
                self.attach(widget, col, row, 2, 1)
            else:
                self.attach(widget, col, row, 1, 1)

    def clear(self):
        for i in self.get_children():
            self.remove(i)



import logging

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import GLib


class Printer:
    def __init__(self, state_cb, state_callbacks):
        self.config = {}
        self.data = {}
        self.state = "disconnected"
        self.state_cb = state_cb
        self.state_callbacks = state_callbacks
        self.power_devices = {}
        self.tools = []
        self.extrudercount = 0
        self.tempdevcount = 0
        self.fancount = 0
        self.ledcount = 0
        self.pwm_tools_count = 0
        self.output_pin_count = 0
        self.store_timeout = None
        self.tempstore = {}
        self.tempstore_size = 1200
        self.cameras = []
        self.available_commands = {}
        self.spoolman = False
        self.active_spool_id = None
        self.active_spool = None
        self.active_spool_checked = False
        self.temp_devices = self.sensors = None
        self.system_info = {}
        self.warnings = []

    def reinit(self, printer_info, data):
        self.config = data['configfile']['config']
        self.data = data
        self.tools.clear()
        self.extrudercount = 0
        self.tempdevcount = 0
        self.fancount = 0
        self.ledcount = 0
        self.output_pin_count = 0
        self.pwm_tools_count = 0
        self.tempstore.clear()
        self.tempstore_size = 1200
        self.available_commands.clear()
        self.temp_devices = self.sensors = None
        self.stop_tempstore_updates()
        self.system_info.clear()
        self.warnings = []
        self.active_spool_id = None
        self.active_spool = None
        self.active_spool_checked = False

        for x in self.config.keys():
            # Support for hiding devices by name
            section, _, name = x.partition(" ")
            if name.startswith("_"):
                continue

            if section.startswith("extruder"):
                self.tools.append(x)
                self.extrudercount += 1
                if name.startswith("extruder_stepper"):
                    continue
                self.data[x] = {"temperature": 0, "target": 0}
            elif section in (
                "heater_bed",
                "heater_generic",
                "temperature_sensor",
                "temperature_fan"
            ):
                self.data[x] = {"temperature": 0}
                if section != "temperature_sensor":
                    self.data[x]["target"] = 0
                self.tempdevcount += 1
            elif section in (
                "fan",
                "controller_fan",
                "heater_fan",
                "fan_generic"
            ):
                self.fancount += 1
            elif section == "output_pin":
                self.output_pin_count += 1
            elif section == "pwm_tool":
                self.pwm_tools_count += 1
            elif section in (
                "led",
                "neopixel",
                "dotstar",
                "pca9533",
                "pca9632"
            ):
                self.ledcount += 1

        self.tools = sorted(self.tools)
        self.log_counts(printer_info)
        self.process_update(data)

    def register_dynamic_sensors(self, objects):
        for obj in objects:
            if obj.startswith("temperature_sensor "):
                name = obj.split(" ", 1)[1]
                if name.startswith("_"):
                    continue
                if obj not in self.config:
                    logging.info(f"Registering dynamic sensor: {obj}")
                    self.config[obj] = {}
                    self.data[obj] = {"temperature": 0}
                    self.tempdevcount += 1

    def log_counts(self, printer_info):
        logging.info(f"Klipper version: {printer_info['software_version']}")
        logging.info(f"# Extruders: {self.extrudercount}")
        logging.info(f"# Temperature devices: {self.tempdevcount}")
        logging.info(f"# Fans: {self.fancount}")
        logging.info(f"# Output pins: {self.output_pin_count}")
        logging.info(f"# PWM tools: {self.pwm_tools_count}")
        logging.info(f"# Leds: {self.ledcount}")

    def stop_tempstore_updates(self):
        if self.store_timeout is not None:
            GLib.source_remove(self.store_timeout)
            self.store_timeout = None

    def process_update(self, data):
        if self.data is None:
            return

        for x in data:
            if x == "configfile":
                if 'config' in data[x]:
                    self.config.update(data[x]['config'])
                if 'warnings' in data[x]:
                    self.warnings = data[x]['warnings']
            if x not in self.data:
                self.data[x] = {}
            self.data[x].update(data[x])

        if "webhooks" in data or "print_stats" in data or "idle_timeout" in data:
            self.process_status_update()

    def evaluate_state(self):
        # webhooks states: startup, ready, shutdown, error
        # print_stats: standby, printing, paused, error, complete
        # idle_timeout: Idle, Printing, Ready
        if self.data['webhooks']['state'] == "ready" and (
                'print_stats' in self.data and 'state' in self.data['print_stats']):
            if self.data['print_stats']['state'] == 'paused':
                return "paused"
            if self.data['print_stats']['state'] == 'printing':
                return "printing"
        return self.data['webhooks']['state']

    def process_status_update(self):
        state = self.evaluate_state()
        if state != self.state:
            self.change_state(state)
        return False

    def process_power_update(self, data):
        if data['device'] in self.power_devices:
            self.power_devices[data['device']]['status'] = data['status']

    def change_state(self, state):
        if state not in list(self.state_callbacks):
            return  # disconnected, startup, ready, shutdown, error, paused, printing
        if state != self.state:
            logging.debug(f"Changing state from '{self.state}' to '{state}'")
            self.state = state
        if self.state_callbacks[state] is not None:
            logging.debug(f"Adding callback for state: {state}")
            GLib.idle_add(self.state_cb, state, self.state_callbacks[state])

    def configure_power_devices(self, data):
        self.power_devices = {}

        logging.debug(f"Processing power devices: {data}")
        for x in data['devices']:
            self.power_devices[x['device']] = {
                "status": "on" if x['status'] == "on" else "off"
            }
        logging.debug(f"Power devices: {self.power_devices}")

    def configure_cameras(self, data):
        self.cameras = data
        logging.debug(f"Cameras: {self.cameras}")

    def get_config_section_list(self, search=""):
        if self.config is not None:
            return [i for i in list(self.config) if i.startswith(search)] if hasattr(self, "config") else []
        return []

    def get_config_section(self, section):
        return self.config[section] if section in self.config else False

    def get_macro(self, macro):
        return next(
            (
                self.config[key]
                for key in self.config.keys()
                if key.find(macro) > -1
            ),
            False,
        )

    def get_fans(self):
        fans = []
        if self.config_section_exists("fan"):
            fans.append("fan")
        for fan_type in ["controller_fan", "fan_generic", "heater_fan"]:
            fans.extend(iter(self.get_config_section_list(f"{fan_type} ")))
        return fans

    def get_pwm_tools(self):
        return self.get_config_section_list("pwm_tool ")

    def get_output_pins(self):
        return self.get_config_section_list("output_pin ")

    def get_gcode_macros(self):
        macros = []
        for macro in self.get_config_section_list("gcode_macro "):
            macro = macro[12:].strip()
            if macro.startswith("_") or macro.upper() in ('LOAD_FILAMENT', 'UNLOAD_FILAMENT'):
                continue
            if self.get_macro(macro) and "rename_existing" in self.get_macro(macro):
                continue
            macros.append(macro)
        return macros

    def get_heaters(self):
        heaters = self.get_config_section_list("heater_generic ")
        if "heater_bed" in self.config:
            heaters.insert(0, "heater_bed")
        return heaters

    def get_temp_fans(self):
        return self.get_config_section_list("temperature_fan")

    def get_temp_sensors(self):
        return self.get_config_section_list("temperature_sensor")

    def get_filament_sensors(self):
        if self.sensors is None:
            self.sensors = list(self.get_config_section_list("filament_switch_sensor "))
            self.sensors.extend(iter(self.get_config_section_list("filament_motion_sensor ")))
        return self.sensors

    def get_probe(self):
        probe_types = ["probe", "bltouch", "smart_effector", "probe_eddy_current"]
        for probe_type in probe_types:
            if self.config_section_exists(probe_type):
                logging.info(f"Probe type: {probe_type}")
                return self.get_config_section(probe_type)
        return None

    def get_printer_status_data(self):
        return {
            "moonraker": {
                "power_devices": {"count": len(self.get_power_devices())},
                "cameras": {"count": len(self.cameras)},
                "spoolman": self.spoolman,
            },
            "printer": {
                "pause_resume": {"is_paused": self.state == "paused"},
                "extruders": {"count": self.extrudercount},
                "temperature_devices": {"count": self.tempdevcount},
                "fans": {"count": self.fancount},
                "output_pins": {"count": self.output_pin_count},
                "pwm_tools": {"count": self.pwm_tools_count},
                "gcode_macros": {"count": len(self.get_gcode_macros()), "list": self.get_gcode_macros()},
                "leds": {"count": self.ledcount},
                "config_sections": list(self.config.keys()),
                "available_commands": self.available_commands,
            }
        }

    def get_leds(self):
        return [
            led
            for led_type in ["dotstar", "led", "neopixel", "pca9533", "pca9632"]
            for led in self.get_config_section_list(f"{led_type} ")
            if not led.split()[1].startswith("_")
        ]

    def get_led_color_order(self, led):
        if led not in self.config or led not in self.data:
            logging.debug(f"Error getting {led} config")
            return None
        elif "color_order" in self.config[led]:
            return self.config[led]["color_order"]
        colors = ''
        for option in self.config[led]:
            if option in ("red_pin", 'initial_red') and 'R' not in colors:
                colors += 'R'
            elif option in ("green_pin", 'initial_green') and 'G' not in colors:
                colors += 'G'
            elif option in ("blue_pin", 'initial_blue') and 'B' not in colors:
                colors += 'B'
            elif option in ("white_pin", 'initial_white') and 'W' not in colors:
                colors += 'W'
        logging.debug(f"Colors in led: {colors}")
        return colors

    def get_power_devices(self):
        return list(self.power_devices)

    def get_power_device_status(self, device):
        if device not in self.power_devices:
            return
        return self.power_devices[device]['status']

    def get_stat(self, stat, substat=None):
        if self.data is None or stat not in self.data:
            return {}
        if substat is not None:
            return self.data.get(stat, {}).get(substat, {})
        else:
            return self.data.get(stat, {})

    def set_stat(self, stat, data):
        if self.data is None:
            logging.error("Data is not initialized")
            return
        if stat not in self.data:
            logging.error(f"Stat '{stat}' not found in data")
            return
        self.data[stat].update(data)

    def get_fan_speed(self, fan="fan"):
        speed = 0
        if fan not in self.config or fan not in self.data:
            logging.debug(f"Error getting {fan} config")
            return speed
        if "speed" in self.data[fan]:
            speed = self.data[fan]["speed"]
        if 'max_power' in self.config[fan]:
            max_power = float(self.config[fan]['max_power'])
            if max_power > 0:
                speed = speed / max_power
        if 'off_below' in self.config[fan]:
            off_below = float(self.config[fan]['off_below'])
            if speed < off_below:
                speed = 0
        return speed

    def get_pin_value(self, pin):
        if pin in self.data:
            return self.data[pin]["value"]
        elif pin in self.config and 'value' in self.config[pin]:
            return self.config[pin]["value"]
        return 0

    def get_temp_store_devices(self):
        return list(self.tempstore)

    def device_has_target(self, device):
        return device in self.data and "target" in self.data[device]

    def device_has_power(self, device):
        return device in self.data and "power" in self.data[device]

    def get_temp_store(self, device, section=False, results=0):
        if device not in self.tempstore:
            return False

        if section is not False:
            if section not in self.tempstore[device]:
                return False
            if results == 0 or results >= len(self.tempstore[device][section]):
                return self.tempstore[device][section]
            return self.tempstore[device][section][-results:]

        temp = {}
        for section in self.tempstore[device]:
            if results == 0 or results >= len(self.tempstore[device][section]):
                temp[section] = self.tempstore[device][section]
            temp[section] = self.tempstore[device][section][-results:]
        return temp

    def get_tempstore_size(self):
        return self.tempstore_size

    def get_temp_devices(self):
        if self.temp_devices is None:
            devices = [
                device
                for device in self.tools
                if not device.startswith('extruder_stepper')
            ]
            self.temp_devices = devices + self.get_heaters() + self.get_temp_sensors() + self.get_temp_fans()
        return self.temp_devices

    def get_tools(self):
        return self.tools

    def get_tool_number(self, tool):
        return self.tools.index(tool)

    def init_temp_store(self, tempstore):
        if self.tempstore and set(self.tempstore) != set(tempstore):
            logging.debug("Tempstore has changed")
            self.tempstore = tempstore
            self.change_state(self.state)
        else:
            self.tempstore = tempstore
        for device in self.tempstore:
            for x in self.tempstore[device]:
                length = len(self.tempstore[device][x])
                if length < self.tempstore_size:
                    for _ in range(1, self.tempstore_size - length):
                        self.tempstore[device][x].insert(0, 0)
        logging.info(f"Temp store: {list(self.tempstore)}")
        if not self.store_timeout:
            self.store_timeout = GLib.timeout_add_seconds(1, self._update_temp_store)

    def config_section_exists(self, section):
        return section in self.get_config_section_list()

    def _update_temp_store(self):
        if self.tempstore is None:
            return False
        for device in self.tempstore:
            for x in self.tempstore[device]:
                self.tempstore[device][x].pop(0)
                temp = self.get_stat(device, x[:-1])
                if not temp:
                    # If the temperature is not available, set it to 0.
                    temp = 0
                self.tempstore[device][x].append(temp)
        return True

    def enable_spoolman(self):
        logging.info("Enabling Spoolman")
        self.spoolman = True

    def set_active_spool(self, spool_id=None, spool_data=None):
        self.active_spool_id = spool_id
        self.active_spool = spool_data



import logging


class SpoolmanAPI:
    """Handles communication with the Spoolman proxy."""

    def __init__(self, api_client):
        self.api = api_client

    def _make_request(self, method="GET", path="/v1/spool", json_body=None):
        """Helper to standardize API calls through the KlipperScreen proxy."""
        try:
            payload = {
                "request_method": method,
                "path": path,
            }
            if json_body:
                payload["body"] = json_body

            result = self.api.post_request("server/spoolman/proxy", json=payload)

            if not result or "result" not in result:
                logging.warning(f"Spoolman API Error: {result}")
                return None

            if isinstance(result["result"], bool):
                return result["result"]

            return result["result"]

        except Exception as e:
            logging.error(f"Spoolman API Exception: {e}")
            return None

    def get_active_spool_id(self) -> int:
        """Fetches the current active spool ID."""
        result = self.api.send_request("server/spoolman/spool_id")
        if not result or "spool_id" not in result:
            return None
        return result["spool_id"]

    def set_active_spool_id(self, spool_id: int) -> bool:
        """Sets the active spool ID."""
        try:
            self.api.post_request("server/spoolman/spool_id", json={"spool_id": spool_id})
            return True
        except Exception as e:
            logging.error(f"Error setting active spool: {e}")
            return False

    def clear_active_spool(self) -> bool:
        """Clears the active spool."""
        try:
            self.api.post_request("server/spoolman/spool_id", json={})
            return True
        except Exception as e:
            logging.error(f"Error clearing active spool: {e}")
            return False

    def get_spool_details(self, spool_id: int) -> dict:
        """Fetches full details for a specific spool."""
        return self._make_request(method="GET", path=f"/v1/spool/{spool_id}")

    def load_all_spools(self, allow_archived: bool = False) -> list:
        """Fetches the full list of spools."""
        path = f"/v1/spool?allow_archived={str(allow_archived).lower()}"
        return self._make_request(method="GET", path=path)

    def update_spool(self, spool_id: int, payload: dict) -> dict:
        """Updates a spool"""
        path = f"/v1/spool/{spool_id}"
        return self._make_request(method="PATCH", path=path, json_body=payload)



import datetime
import logging

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk
from ks_includes.screen_panel import ScreenPanel


class Panel(ScreenPanel):
    def __init__(self, screen, title):
        title = title or _("System")
        super().__init__(screen, title)
        self.current_row = 0
        self.mem_multiplier = None
        self.scales = {}
        self.labels = {}
        self.grid = Gtk.Grid(column_spacing=10, row_spacing=5)

        self.sysinfo = screen.printer.system_info
        if not self.sysinfo:
            logging.debug("Asking for info")
            self.sysinfo = screen.apiclient.send_request("machine/system_info")
            if 'system_info' in self.sysinfo:
                screen.printer.system_info = self.sysinfo['system_info']
                self.sysinfo = self.sysinfo['system_info']
        logging.debug(self.sysinfo)
        if self.sysinfo:
            self.content.add(self.create_layout())
        else:
            self.content.add(Gtk.Label(label=_("No info available"), vexpand=True))

    def back(self):
        if not self.sysinfo:
            self._screen.panels_reinit.append("system")
        return False

    def create_layout(self):
        self.cpu_count = int(self.sysinfo["cpu_info"]["cpu_count"])
        self.labels["cpu_usage"] = Gtk.Label(label="", xalign=0)
        self.grid.attach(self.labels["cpu_usage"], 0, self.current_row, 1, 1)
        self.scales["cpu_usage"] = Gtk.ProgressBar(
            hexpand=True, show_text=False, fraction=0
        )
        self.grid.attach(self.scales["cpu_usage"], 1, self.current_row, 1, 1)
        self.current_row += 1

        for i in range(self.cpu_count):
            self.labels[f"cpu_usage_{i}"] = Gtk.Label(label="", xalign=0)
            self.grid.attach(self.labels[f"cpu_usage_{i}"], 0, self.current_row, 1, 1)
            self.scales[f"cpu_usage_{i}"] = Gtk.ProgressBar(
                hexpand=True, show_text=False, fraction=0
            )
            self.grid.attach(self.scales[f"cpu_usage_{i}"], 1, self.current_row, 1, 1)
            self.current_row += 1

        self.labels["memory_usage"] = Gtk.Label(label="", xalign=0)
        self.grid.attach(self.labels["memory_usage"], 0, self.current_row, 1, 1)
        self.scales["memory_usage"] = Gtk.ProgressBar(
            hexpand=True, show_text=False, fraction=0
        )
        self.grid.attach(self.scales["memory_usage"], 1, self.current_row, 1, 1)
        self.current_row += 1

        self.grid.attach(Gtk.Separator(), 0, self.current_row, 2, 1)
        self.current_row += 1
        self.populate_info()

        scroll = self._gtk.ScrolledWindow()
        scroll.add(self.grid)
        return scroll

    def set_mem_multiplier(self, data):
        memory_units = data.get("memory_units", "kB").lower()
        units_mapping = {
            "kb": 1024,
            "mb": 1024**2,
            "gb": 1024**3,
            "tb": 1024**4,
            "pb": 1024**5,
        }
        self.mem_multiplier = units_mapping.get(memory_units, 1)

    def add_label_to_grid(self, text, column, bold=False):
        if bold:
            text = f"<b>{text}</b>"
        label = Gtk.Label(label=text, use_markup=True, xalign=0, wrap=True)
        self.grid.attach(label, column, self.current_row, 1, 1)
        self.current_row += 1

    def populate_info(self):
        self.add_label_to_grid(self.prettify("date"), 0, bold=True)
        self.labels["date"] = Gtk.Label(label="", xalign=0)
        self.grid.attach(self.labels["date"], 1, self.current_row - 1, 1, 1)
        self.add_label_to_grid("", 0)

        for category, data in self.sysinfo.items():
            if category == "python":
                self.add_label_to_grid(self.prettify(category), 0, bold=True)
                self.current_row -= 1
                self.add_label_to_grid(
                    f'Version: {data["version_string"].split(" ")[0]}', 1
                )
                continue

            if (
                category
                in (
                    "virtualization",
                    "provider",
                    "available_services",
                    "service_state",
                    "instance_ids",
                )
                or not self.sysinfo[category]
            ):
                continue

            self.add_label_to_grid(self.prettify(category), 0, bold=True)

            if isinstance(data, dict):
                for key, value in data.items():
                    if key in ("version_parts", "memory_units") or not value:
                        continue
                    if key == "total_memory":
                        if not self.mem_multiplier:
                            self.set_mem_multiplier(data)
                        value = self.format_size(int(value) * self.mem_multiplier)
                    if isinstance(value, dict):
                        self.add_label_to_grid(self.prettify(key), 0)
                        self.current_row -= 1
                        for sub_key, sub_value in value.items():
                            if not sub_value:
                                continue
                            elif (
                                isinstance(sub_value, list)
                                and sub_key == "ip_addresses"
                            ):
                                for _ip in sub_value:
                                    self.add_label_to_grid(
                                        f"{_('IP Address')}: {_ip['address']}", 1
                                    )
                                continue
                            self.add_label_to_grid(
                                f"{self.prettify(sub_key)}: {sub_value}", 1
                            )
                    else:
                        self.add_label_to_grid(f"{self.prettify(key)}: {value}", 1)
            # Add empty line
            self.add_label_to_grid("", 0)

    def process_update(self, action, data):
        if not self.sysinfo:
            return
        if action == "notify_proc_stat_update":
            self.labels["cpu_usage"].set_label(
                f'CPU: {data["system_cpu_usage"]["cpu"]:.0f}%'
            )
            self.scales["cpu_usage"].set_fraction(
                float(data["system_cpu_usage"]["cpu"]) / 100
            )
            for i in range(self.cpu_count):
                self.labels[f"cpu_usage_{i}"].set_label(
                    f'CPU {i}: {data["system_cpu_usage"][f"cpu{i}"]:.0f}%'
                )
                self.scales[f"cpu_usage_{i}"].set_fraction(
                    float(data["system_cpu_usage"][f"cpu{i}"]) / 100
                )

            self.labels["memory_usage"].set_label(
                _("Memory")
                + f': {(data["system_memory"]["used"] / data["system_memory"]["total"]) * 100:.0f}%'
            )
            self.scales["memory_usage"].set_fraction(
                float(data["system_memory"]["used"])
                / float(data["system_memory"]["total"])
            )
            now = datetime.datetime.now().astimezone()
            self.labels["date"].set_label(now.strftime("%a %b %e %H:%M:%S %Z %Y"))



#!/usr/bin/python

import ast
import argparse
import gc
import json
import logging
import os
import subprocess
import pathlib
import traceback  # noqa
import locale
import re
import sys
import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, Gdk, GLib, Pango
from importlib import import_module
from jinja2 import Environment
from signal import SIGTERM
from datetime import datetime

from ks_includes import functions
from ks_includes.KlippyWebsocket import KlippyWebsocket
from ks_includes.KlippyRest import KlippyRest
from ks_includes.files import KlippyFiles
from ks_includes.KlippyGtk import KlippyGtk
from ks_includes.printer import Printer
from ks_includes.spoolman_api import SpoolmanAPI
from ks_includes.widgets.keyboard import Keyboard
from ks_includes.widgets.prompts import Prompt
from ks_includes.widgets.lockscreen import LockScreen
from ks_includes.widgets.screensaver import ScreenSaver
from ks_includes.config import KlipperScreenConfig
from panels.base_panel import BasePanel


logging.getLogger("urllib3").setLevel(logging.WARNING)

klipperscreendir = pathlib.Path(__file__).parent.resolve()


def set_text_direction(lang=None):
    rtl_languages = ['he']
    if lang is None:
        for lng in rtl_languages:
            if locale.getlocale()[0].startswith(lng):
                lang = lng
                break
    if lang in rtl_languages:
        Gtk.Widget.set_default_direction(Gtk.TextDirection.RTL)
        logging.debug("Enabling RTL mode")
        return False
    Gtk.Widget.set_default_direction(Gtk.TextDirection.LTR)
    return True


class KlipperScreen(Gtk.Window):
    _cur_panels = []
    connecting = False
    connecting_to_printer = None
    connected_printer = None
    files = None
    keyboard = None
    keyboard_cache = {}
    panels = {}
    popup_message = None
    printers = None
    printer = None
    updating = False
    _ws = None
    reinit_count = 0
    max_retries = 4
    initialized = False
    initializing = False
    popup_timeout = None
    wayland = False
    windowed = False
    notification_log = []
    prompt = None
    tempstore_timeout = None
    check_dpms_timeout = None

    def __init__(self, args):
        self.server_info = None
        try:
            super().__init__(title="KlipperScreen")
        except Exception as e:
            logging.exception(f"{e}\n\n{traceback.format_exc()}")
            raise RuntimeError from e
        GLib.set_prgname('KlipperScreen')
        self.blanking_time = 600
        self.apiclient = None
        self.dialogs = []
        self.confirm = None
        self.panels_reinit = []
        self.last_popup_time = datetime.now()

        configfile = os.path.normpath(os.path.expanduser(args.configfile))

        self._config = KlipperScreenConfig(configfile, self)
        self.env = Environment(extensions=["jinja2.ext.i18n"], autoescape=True)
        self.env.install_gettext_translations(self._config.get_lang())

        self.connect("key-press-event", self._key_press_event)
        self.connect("configure_event", self.update_size)
        display = Gdk.Display.get_default()
        self.display_number = os.environ.get('DISPLAY') or ':0'
        logging.debug(f"Display for xset: {self.display_number}")
        monitor_amount = Gdk.Display.get_n_monitors(display)
        if (monitor_amount):
            for i in range(monitor_amount):
                m = display.get_monitor(i)
                logging.info(f"Screen {i}: {m.get_geometry().width}x{m.get_geometry().height}")
        else:
            logging.warning("WARNING: No monitors detected by Gdk")
        try:
            mon_n = int(args.monitor)
            if not (-1 < mon_n < monitor_amount):
                raise ValueError
        except ValueError:
            mon_n = 0
        logging.info(f"Monitors: {monitor_amount} using number: {mon_n}")
        monitor = display.get_monitor(mon_n)
        self.wayland = display.get_name().startswith('wayland') or display.get_primary_monitor() is None
        logging.info(f"Wayland: {self.wayland} Display name: {display.get_name()}")
        self.width = self._config.get_main_config().getint("width", None)
        self.height = self._config.get_main_config().getint("height", None)
        if 'XDG_CURRENT_DESKTOP' in os.environ:
            logging.warning("Running inside a desktop environment is not recommended")
            if not self.width:
                self.width = max(int(monitor.get_geometry().width * .5), 480)
            if not self.height:
                self.height = max(int(monitor.get_geometry().height * .5), 320)
        if self.width or self.height:
            logging.info("Setting windowed mode")
            if mon_n > 0:
                logging.error("Monitor selection is only supported for fullscreen")
            self.set_resizable(True)
            self.windowed = True
        else:
            self.width = monitor.get_geometry().width
            self.height = monitor.get_geometry().height
            self.fullscreen_on_monitor(self.get_screen(), mon_n)
        self.set_default_size(self.width, self.height)
        self.aspect_ratio = self.width / self.height
        self.vertical_mode = self.aspect_ratio < 1.0
        logging.info(f"Screen resolution: {self.width}x{self.height}")
        self.theme = self._config.get_main_config().get('theme')
        self.show_cursor = self._config.get_main_config().getboolean("show_cursor", fallback=False)
        self.setup_gtk_settings()
        self.style_provider = Gtk.CssProvider()
        self.screensaver = ScreenSaver(self)
        self.lock_screen = LockScreen(self)
        self.gtk = KlippyGtk(self)
        self.base_css = ""
        self.load_base_styles()
        self.set_icon_from_file(os.path.join(klipperscreendir, "styles", "icon.svg"))
        self.base_panel = BasePanel(self)
        self.change_theme(self.theme)
        self.overlay = Gtk.Overlay()
        self.add(self.overlay)
        self.overlay.add_overlay(self.base_panel.main_grid)
        self.show_all()
        self.update_cursor(self.show_cursor)
        min_ver = (3, 8)
        if sys.version_info < min_ver:
            self.show_error_modal(
                "Error",
                _("The system doesn't meet the minimum requirement") + "\n"
                + _("Minimum:") + f" Python {min_ver[0]}.{min_ver[1]}" + "\n"
                + _("System:") + f" Python {sys.version_info.major}.{sys.version_info.minor}"
            )
            return
        if self._config.errors:
            self.show_error_modal("Invalid config file", self._config.get_errors())
            return
        self.base_panel.activate()
        self.use_dpms = self._config.get_main_config().getboolean("use_dpms", fallback=(not self.wayland))
        self.use_dpms &= functions.dpms_loaded
        self.set_dpms(self.use_dpms)
        autolock = self._config.get_main_config().getint("autolock_timeout", fallback=0)
        self.lock_screen.set_autolock_timeout(autolock)
        self.log_notification("KlipperScreen Started", 1)
        self.initial_connection()
        if self._config.get_main_config().getboolean('start_locked', False):
            self.lock_screen.lock(None)

    def update_cursor(self, show: bool):
        self.show_cursor = show
        self.gtk.set_cursor(show, window=self.get_window())

    def state_execute(self, state, callback):
        self.screensaver.close()
        if 'printer_select' in self._cur_panels:
            logging.debug(f"Connected printer changed {state}")
            return False
        if state in ("printing", "paused"):
            self.set_screenblanking_timeout(self._config.get_main_config().get('screen_blanking_printing'))
        else:
            self.set_screenblanking_timeout(self._config.get_main_config().get('screen_blanking'))
            for warning in self.printer.warnings:
                self.show_popup_message(f"Klipper:\n{warning['message']}", level=2)
        callback()
        return False

    def initial_connection(self):
        self.printers = self._config.get_printers()
        state_callbacks = {
            "disconnected": self.state_disconnected,
            "error": self.state_error,
            "paused": self.state_paused,
            "printing": self.state_printing,
            "ready": self.state_ready,
            "startup": self.state_startup,
            "shutdown": self.state_shutdown
        }
        for printer in self.printers:
            printer["data"] = Printer(self.state_execute, state_callbacks)
        default_printer = self._config.get_main_config().get('default_printer')
        logging.debug(f"Default printer: {default_printer}")
        if [True for p in self.printers if default_printer in p]:
            self.connect_printer(default_printer)
        elif len(self.printers) == 1:
            pname = list(self.printers[0])[0]
            self.connect_printer(pname)
        else:
            self.base_panel.show_printer_select(True)
            self.show_printer_select()

    def close_websocket(self):
        self._ws.close()
        self.connected_printer = None
        self.printer.state = "disconnected"

    def connect_printer(self, name):
        self.connecting_to_printer = name
        if self._ws is not None and self._ws.connected:
            self.printer_initializing("Waiting Websocket closure")
            self.close_websocket()
            return
        gc.collect()
        self.connecting = True
        self.initialized = False
        self.initializing = False
        logging.info(f"Connecting to printer: {name}")
        ind = next(
            (
                self.printers.index(printer)
                for printer in self.printers
                if name == list(printer)[0]
            ),
            0,
        )
        self.printer = self.printers[ind]["data"]
        self.apiclient = KlippyRest(
            self.printers[ind][name]["moonraker_host"],
            self.printers[ind][name]["moonraker_port"],
            self.printers[ind][name]["moonraker_api_key"],
            self.printers[ind][name]["moonraker_path"],
            self.printers[ind][name]["moonraker_ssl"],
        )

        self.spoolman_api = SpoolmanAPI(self.apiclient)

        self._ws = KlippyWebsocket(
            {
                "on_connect": self.websocket_connected,
                "on_message": self._websocket_callback,
                "on_close": self.websocket_disconnected,
                "on_cancel": self.websocket_connection_cancel,
            },
            self.printers[ind][name]["moonraker_host"],
            self.printers[ind][name]["moonraker_port"],
            self.printers[ind][name]["moonraker_api_key"],
            self.printers[ind][name]["moonraker_path"],
            self.printers[ind][name]["moonraker_ssl"],
        )
        if self.files is None:
            self.files = KlippyFiles(self)
        else:
            self.files.reinit()

        self.reinit_count = 0
        self.printer_initializing(_("Connecting to %s") % name, True)
        self.connect_to_moonraker()

    def ws_subscribe(self):
        requested_updates = {
            "objects": {
                "bed_mesh": ["profile_name", "mesh_max", "mesh_min", "probed_matrix", "profiles"],
                "configfile": ["config", "warnings"],
                "display_status": ["progress", "message"],
                "fan": ["speed"],
                "gcode_move": ["extrude_factor", "gcode_position", "homing_origin", "speed_factor", "speed"],
                "idle_timeout": ["state"],
                "pause_resume": ["is_paused"],
                "print_stats": ["print_duration", "total_duration", "filament_used", "filename", "state", "message",
                                "info"],
                "toolhead": ["homed_axes", "estimated_print_time", "print_time", "position", "extruder",
                             "max_accel", "minimum_cruise_ratio", "max_velocity", "square_corner_velocity"],
                "virtual_sdcard": ["file_position", "is_active", "progress"],
                "webhooks": ["state", "state_message"],
                "firmware_retraction": ["retract_length", "retract_speed", "unretract_extra_length", "unretract_speed"],
                "motion_report": ["live_position", "live_velocity", "live_extruder_velocity"],
                "exclude_object": ["current_object", "objects", "excluded_objects"],
                "manual_probe": ['is_active'],
                "screws_tilt_adjust": ['results', 'error', 'max_deviation'],
            }
        }
        for extruder in self.printer.get_tools():
            requested_updates['objects'][extruder] = [
                "target", "temperature", "pressure_advance", "smooth_time", "power"]
        for h in self.printer.get_heaters():
            requested_updates['objects'][h] = ["target", "temperature", "power"]
        for t in self.printer.get_temp_sensors():
            requested_updates['objects'][t] = ["temperature"]
        for f in self.printer.get_temp_fans():
            requested_updates['objects'][f] = ["target", "temperature"]
        for f in self.printer.get_fans():
            requested_updates['objects'][f] = ["speed"]
        for f in self.printer.get_filament_sensors():
            requested_updates['objects'][f] = ["enabled", "filament_detected"]
        for p in self.printer.get_pwm_tools() + self.printer.get_output_pins():
            requested_updates['objects'][p] = ["value"]
        for led in self.printer.get_leds():
            requested_updates['objects'][led] = ["color_data"]

        self._ws.klippy.object_subscription(requested_updates)

    @staticmethod
    def _load_panel(panel):
        logging.debug(f"Loading panel: {panel}")
        panel_path = os.path.join(os.path.dirname(__file__), 'panels', f"{panel}.py")
        if not os.path.exists(panel_path):
            logging.error(f"Panel {panel} does not exist")
            raise FileNotFoundError(os.strerror(2), "\n" + panel_path)
        return import_module(f"panels.{panel}")

    def show_panel(self, panel, title=None, remove_all=False, panel_name=None, **kwargs):
        if panel_name is None:
            panel_name = panel
        if panel == "lock_screen":
            self.lock_screen.lock(None)
            return
        if self._cur_panels and panel_name == self._cur_panels[-1]:
            logging.error("Panel is already is in view")
            return
        try:
            if remove_all:
                self.panels_reinit = list(self.panels)
                if panel in self._cur_panels:
                    self._menu_go_back(home=True)
                else:
                    self._remove_all_panels()
                    for dialog in self.dialogs:
                        self.gtk.remove_dialog(dialog)
            else:
                self._remove_current_panel()
            if panel_name not in self.panels:
                try:
                    self.panels[panel_name] = self._load_panel(panel).Panel(self, title, **kwargs)
                except Exception as e:
                    self.show_error_modal(f"Unable to load panel {panel}", f"{e}\n\n{traceback.format_exc()}")
                    return
            elif panel_name in self.panels_reinit:
                logging.info(f"Reinitializing panel {panel}")
                self.panels[panel_name].__init__(self, title, **kwargs)
                self.panels_reinit.remove(panel_name)
            self._cur_panels.append(panel_name)
            if 'extra' in kwargs and hasattr(self.panels[panel], "set_extra"):
                self.panels[panel].set_extra(**kwargs)
            self.attach_panel(panel_name)
        except Exception as e:
            logging.exception(f"Error attaching panel:\n{e}\n\n{traceback.format_exc()}")

    def set_panel_title(self, title):
        self.base_panel.set_title(title)

    def attach_panel(self, panel):
        if panel in self.panels_reinit:
            # this happens when the first panel needs a reinit
            self.reload_panels()
            return
        self.base_panel.add_content(self.panels[panel])
        logging.debug(f"Current panel hierarchy: {' > '.join(self._cur_panels)}")
        while len(self.panels[panel].menu) > 1:
            self.panels[panel].unload_menu()
        if hasattr(self.panels[panel], "process_update"):
            self.process_update("notify_status_update", self.printer.data)
        if hasattr(self.panels[panel], "activate"):
            self.panels[panel].activate()
        self.show_all()

    def log_notification(self, message, level=0):
        time = datetime.now().strftime("%H:%M:%S")
        log_entry = {"message": message, "level": level, "time": time}
        if len(self.notification_log) > 999:
            del self.notification_log[0]
        self.notification_log.append(log_entry)
        self.process_update("notify_log", log_entry)

    def notification_log_clear(self):
        self.notification_log.clear()

    def show_popup_message(self, message, level=3, from_ws=False):
        if from_ws:
            if (datetime.now() - self.last_popup_time).seconds < 1:
                return
            self.last_popup_time = datetime.now()

        self.screensaver.close()
        if self.popup_message is not None:
            self.close_popup_message()

        self.log_notification(message, level)

        msg = Gtk.Button(label=f"{message}", hexpand=True, vexpand=True)
        for widget in msg.get_children():
            if isinstance(widget, Gtk.Label):
                widget.set_line_wrap(True)
                widget.set_line_wrap_mode(Pango.WrapMode.WORD_CHAR)
                widget.set_max_width_chars(40)
        msg.connect("clicked", self.close_popup_message)
        msg.get_style_context().add_class("message_popup")
        if level == 1:
            msg.get_style_context().add_class("message_popup_echo")
            logging.info(f'echo: {message}')
        elif level == 2:
            msg.get_style_context().add_class("message_popup_warning")
            logging.info(f'warning: {message}')
        else:
            msg.get_style_context().add_class("message_popup_error")
            logging.info(f'error: {message}')

        popup = Gtk.Popover(relative_to=self.base_panel.titlebar,
                            halign=Gtk.Align.CENTER, width_request=int(self.width * .9))
        popup.get_style_context().add_class("message_popup_popover")
        popup.add(msg)
        popup.popup()

        self.popup_message = popup
        self.popup_message.show_all()

        if self._config.get_main_config().getboolean('autoclose_popups', True):
            if self.popup_timeout is not None:
                GLib.source_remove(self.popup_timeout)
                self.popup_timeout = None
            timeout = 300 if level == 2 else 10
            self.popup_timeout = GLib.timeout_add_seconds(timeout, self.close_popup_message)

        return False

    def close_popup_message(self, widget=None):
        if self.popup_message is None:
            return False
        self.popup_message.popdown()
        if self.popup_timeout is not None:
            GLib.source_remove(self.popup_timeout)
            self.popup_timeout = None
        self.popup_message = None
        return False

    def show_error_modal(self, title_msg, description="", help_msg=None):
        logging.error(f"Showing error modal: {title_msg} {description}")

        title = Gtk.Label(wrap=True, wrap_mode=Pango.WrapMode.CHAR, hexpand=True, halign=Gtk.Align.START)
        title.set_markup(f"<b>{title_msg}</b>\n")
        version = Gtk.Label(label=f"{functions.get_software_version()}", halign=Gtk.Align.END)

        if not help_msg:
            help_msg = _("Provide KlipperScreen.log when asking for help.\n")
        message = Gtk.Label(label=f"{description}\n\n{help_msg}", wrap=True, wrap_mode=Pango.WrapMode.CHAR)
        scroll = self.gtk.ScrolledWindow()
        scroll.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        scroll.add(message)

        grid = Gtk.Grid()
        grid.attach(title, 0, 0, 1, 1)
        grid.attach(version, 1, 0, 1, 1)
        grid.attach(Gtk.Separator(), 0, 1, 2, 1)
        grid.attach(scroll, 0, 2, 2, 1)

        buttons = [
            {"name": _("Close"), "response": Gtk.ResponseType.CLOSE}
        ]
        self.gtk.Dialog(_("Error"), buttons, grid, self.error_modal_response)

    @staticmethod
    def error_modal_response(dialog, response_id):
        os._exit(1)

    def restart_ks(self, *args):
        logging.debug(f"Restarting {sys.executable} {' '.join(sys.argv)}")
        os.execv(sys.executable, ['python'] + sys.argv)
        # noinspection PyUnreachableCode
        self._ws.send_method("machine.services.restart", {"service": "KlipperScreen"})  # Fallback

    def setup_gtk_settings(self):
        settings = Gtk.Settings.get_default()
        settings.set_property("gtk-theme-name", "Adwaita")
        settings.set_property("gtk-application-prefer-dark-theme", False)

    def load_base_styles(self):
        base_conf_path = os.path.join(klipperscreendir, "styles", "base.conf")
        with open(base_conf_path) as f:
            self.style_options = json.load(f)
        self.gtk.color_list = self.style_options['graph_colors']
        base_css_path = os.path.join(klipperscreendir, "styles", "base.css")
        self.base_css = pathlib.Path(base_css_path).read_text()
        self.base_css = self.base_css.replace("KS_FONT_SIZE", f"{self.gtk.font_size}")
        self.base_css = self.customize_graph_colors(self.base_css)

        self.style_provider.load_from_data(self.base_css.encode())
        Gtk.StyleContext.add_provider_for_screen(
            Gdk.Screen.get_default(),
            Gtk.CssProvider(),
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
        )

    def load_custom_theme(self, theme_name):
        theme_dir = os.path.join(klipperscreendir, "styles", theme_name)
        theme_css_path = os.path.join(theme_dir, "style.css")
        theme_conf_path = os.path.join(theme_dir, "style.conf")

        theme_css = ""
        theme_options = {}

        if os.path.exists(theme_css_path):
            theme_css = pathlib.Path(theme_css_path).read_text()

        if os.path.exists(theme_conf_path):
            try:
                with open(theme_conf_path) as f:
                    theme_options = json.load(f)
            except Exception as e:
                logging.error(
                    f"Unable to parse custom template conf file:\n"
                    f"{e}\n\n"
                    f"{traceback.format_exc()}"
                )

        return theme_css, theme_options

    def customize_graph_colors(self, css_data):
        for category, category_data in self.style_options['graph_colors'].items():
            for i, color in enumerate(category_data['colors'], start=0):
                if category == "extruder":
                    class_name = f".graph_label_{category}{i}" if i > 0 else f".graph_label_{category}"
                elif category == "bed":
                    class_name = f".graph_label_heater_{category}"
                else:
                    class_name = f".graph_label_{category}_{i + 1}"
                css_data += f"\n{class_name} {{ border-left-color: #{color} }}"
        return css_data

    def update_style_provider(self, theme_css):
        css_data = self.customize_graph_colors(theme_css)
        css_data = self.base_css + css_data
        screen = Gdk.Screen.get_default()
        if self.style_provider:
            Gtk.StyleContext.remove_provider_for_screen(screen, self.style_provider)
        self.style_provider = Gtk.CssProvider()
        self.style_provider.load_from_data(css_data.encode())
        Gtk.StyleContext.add_provider_for_screen(
            screen,
            self.style_provider,
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
        )

    def change_theme(self, theme_name=None):
        if not theme_name:
            theme_name = self._config.get_main_config().get('theme')
        self.gtk.update_themedir(theme_name)
        theme_css, theme_options = self.load_custom_theme(theme_name)
        self.style_options.update(theme_options)
        self.gtk.color_list = self.style_options['graph_colors']
        self.update_style_provider(theme_css)
        self.reload_icon_theme()

    def reload_icon_theme(self):
        self.panels_reinit = list(self.panels)
        self.base_panel.reload_icons()

    def _go_to_submenu(self, widget, name):
        logging.info(f"#### Go to submenu {name}")
        # Find current menu item
        if "main_menu" in self._cur_panels:
            menu = "__main"
        elif "splash_screen" in self._cur_panels:
            menu = "__splashscreen"
        else:
            menu = "__print"

        logging.info(f"#### Menu {menu}")
        disname = self._config.get_menu_name(menu, name)
        menuitems = self._config.get_menu_items(menu, name)
        if len(menuitems) != 0:
            self.show_panel("menu", disname, panel_name=name, items=menuitems)
            logging.info(f"menu, {disname}, panel_name={name}, items={menuitems}")
        else:
            logging.info("No items in menu")

    def _remove_all_panels(self):
        logging.debug("Removing all panels")
        while len(self._cur_panels) > 0:
            self._remove_current_panel()
            del self._cur_panels[-1]
        self._cur_panels.clear()
        self.screensaver.close()
        gc.collect()

    def _remove_current_panel(self):
        if not self._cur_panels:
            return
        if hasattr(self.panels[self._cur_panels[-1]], "deactivate"):
            self.panels[self._cur_panels[-1]].deactivate()
        self.base_panel.remove(self.panels[self._cur_panels[-1]].content)

    def _menu_go_back(self, widget=None, home=False):
        logging.info(f"#### Menu go {'home' if home else 'back'}")
        self.remove_keyboard()
        while len(self._cur_panels) > 1:
            self._remove_current_panel()
            del self._cur_panels[-1]
            if not home:
                break
        self.attach_panel(self._cur_panels[-1])

    def check_dpms_state(self):
        if not self.use_dpms:
            return False
        state = functions.get_DPMS_state()
        if state == functions.DPMS_State.Fail:
            self.show_popup_message(_("DPMS has failed and has been disabled"))
            self.set_dpms(False)
            return False
        elif state != functions.DPMS_State.On:
            if not self.screensaver.is_showing():
                self.screensaver.show()
        return True

    def wake_screen(self):
        if self.wayland:
            return
        # Wake the screen (it will go to standby as configured)
        if not self.use_dpms:
            logging.debug("DPMS is disabled cannot wake the screen")
            return
        if self._config.get_main_config().get('screen_blanking') != "off":
            logging.debug("Screen wake up")
        try:
            cmd = ["xset", "-display", self.display_number, "dpms", "force", "on"]
            subprocess.run(cmd, check=True)
        except subprocess.CalledProcessError as e:
            self.show_popup_message(f"Error: {e}")
            self.set_dpms(False)
            return

    def set_dpms(self, use_dpms):
        if self.wayland:
            self.use_dpms = False
            self._config.set("main", "use_dpms", False)
            self._config.save_user_config_options()
            logging.debug("DPMS handling not supported on Wayland")
            return
        if not use_dpms:
            if self.check_dpms_timeout is not None:
                GLib.source_remove(self.check_dpms_timeout)
            self.check_dpms_timeout = None
            state = functions.get_DPMS_state()
            if state != functions.DPMS_State.Fail:
                try:
                    cmd = ["xset", "-display", self.display_number, "dpms", "0", "0", "0"]
                    subprocess.run(cmd, check=True)
                    cmd = ["xset", "-display", self.display_number, "-dpms"]
                    subprocess.run(cmd, check=True)
                except subprocess.CalledProcessError as e:
                    self.show_popup_message(f"FAILED to turn DPMS off on {self.display_number}:\n {e}")
                    return
        self.use_dpms = use_dpms
        self._config.set("main", "use_dpms", use_dpms)
        self._config.save_user_config_options()
        if self.printer and self.printer.state in ("printing", "paused"):
            self.set_screenblanking_timeout(self._config.get_main_config().get('screen_blanking_printing'))
        else:
            self.set_screenblanking_timeout(self._config.get_main_config().get('screen_blanking'))

    def set_dpms_timeout(self):
        try:
            cmd = ["xset", "-display", self.display_number, "dpms", "0", f"{self.blanking_time}", "0"]
            subprocess.run(cmd, check=True)
            logging.info(f"DPMS on {self.display_number} set to: {self.blanking_time}")
        except subprocess.CalledProcessError as e:
            self.show_popup_message(f"DPMS Error:\n {e}")
            self.set_dpms(False)
            return
        if self.blanking_time > 0 and self.check_dpms_timeout is None:
            self.check_dpms_timeout = GLib.timeout_add_seconds(1, self.check_dpms_state)
            return

    def set_autolock_timeout(self, time):
        self.lock_screen.set_autolock_timeout(time)

    def set_screenblanking_printing_timeout(self, time):
        if self.printer and self.printer.state in ("printing", "paused"):
            self.set_screenblanking_timeout(time)

    def set_screenblanking_timeout(self, time):
        # disable screensaver we have our own
        if not self.wayland:
            cmd = ["xset", "-display", self.display_number, "s", "off"]
            subprocess.call(cmd)
            cmd = ["xset", "-display", self.display_number, "s", "noblank"]
            subprocess.call(cmd)
        if time == "off":
            self.blanking_time = 0
        else:
            try:
                self.blanking_time = abs(int(time))
            except Exception as exc:
                logging.exception(exc)
        if self.use_dpms and not self.wayland:
            self.set_dpms_timeout()
        else:
            self.screensaver.reset_timeout()
        logging.debug(f"Blanking timeout: {time}")

    def show_printer_select(self, widget=None):
        self.base_panel.show_heaters(False)
        self.show_panel("printer_select", remove_all=True)

    def websocket_connection_cancel(self):
        self.printer_initializing(
            _("Cannot connect to Moonraker") + '\n\n'
            + f'{self.apiclient.status}'
        )

    def websocket_connected(self):
        logging.debug("### websocket_connected")
        self._ws.klippy.identify_client(functions.get_software_version(), self._ws.api_key)
        self.reinit_count = 0
        self.connecting = False
        self.connected_printer = self.connecting_to_printer
        self.base_panel.set_ks_printer_cfg(self.connected_printer)
        self.init_moonraker_components()
        self.init_klipper()

    def websocket_disconnected(self):
        logging.debug("### websocket_disconnected")
        self.printer.state = "disconnected"
        self.connecting = True
        self.connected_printer = None
        self.initialized = False
        if 'printer_select' not in self._cur_panels:
            self.printer_initializing(_("Lost Connection to Moonraker"), go_to_splash=True)
            self.connect_printer(self.connecting_to_printer)
        else:
            self.panels['printer_select'].disconnected_callback()

    def state_disconnected(self):
        logging.debug("### Going to disconnected")
        self.printer.stop_tempstore_updates()
        self.initialized = False
        self.reinit_count = 0
        self._init_printer(_("Klipper has disconnected"), go_to_splash=True)

    def state_error(self):
        msg = _("Klipper has encountered an error.") + "\n"
        state = self.printer.get_stat("webhooks", "state_message")
        if "FIRMWARE_RESTART" in state:
            msg += _("A FIRMWARE_RESTART may fix the issue.") + "\n"
        elif "micro-controller" in state:
            msg += _("Please recompile and flash the micro-controller.") + "\n"
        self.printer_initializing(msg + "\n" + state, go_to_splash=True)

    def state_paused(self):
        self.state_printing()
        if self._config.get_main_config().getboolean("auto_open_extrude", fallback=True):
            self.show_panel("extrude")

    def state_printing(self):
        self.show_panel("job_status", remove_all=True)

    def state_ready(self, wait=True):
        # Do not return to main menu if completing a job, timeouts/user input will return
        if "job_status" in self._cur_panels and wait:
            return
        if not self.initialized:
            logging.debug("Printer not initialized yet")
            self.printer.state = "not ready"
            return
        self.files.refresh_files()
        self.show_panel("main_menu", remove_all=True, items=self._config.get_menu_items("__main"))

    def state_startup(self):
        self.printer_initializing(_("Klipper is attempting to start"))

    def state_shutdown(self):
        self.printer.stop_tempstore_updates()
        msg = self.printer.get_stat("webhooks", "state_message")
        self.printer_initializing(_("Klipper has shutdown") + "\n\n" + msg, go_to_splash=True)

    def toggle_shortcut(self, show):
        if show and not self.printer.get_printer_status_data()["printer"]["gcode_macros"]["count"] > 0:
            self.show_popup_message(
                _("No eligible macros:") + "\n"
                + _("macros with a name starting with '_' are hidden") + "\n"
                + _("macros that use 'rename_existing' are hidden") + "\n"
                + _("LOAD_FILAMENT/UNLOAD_FILAMENT are hidden and should be used from extrude") + "\n"
            )
        self.base_panel.show_shortcut(show)

    def change_language(self, widget, lang):
        self._config.install_language(lang)
        self.env.install_gettext_translations(self._config.get_lang())
        self._config._create_configurable_options(self)
        self._config.set('main', 'language', lang)
        self._config.save_user_config_options()
        self.reload_panels()

    def reload_panels(self, *args):
        if "printer_select" in self._cur_panels:
            self.show_printer_select()
            return
        home = self._cur_panels[0]
        self.panels_reinit = list(self.panels)
        self._remove_all_panels()
        if home == "main_menu":
            self.show_panel(home, items=self._config.get_menu_items("__main"))
        else:
            self.show_panel(home)

    def _websocket_callback(self, action, data):
        if self.connecting:
            logging.debug("Not connected")
            return
        if action == "notify_klippy_disconnected":
            self.printer.process_update({'webhooks': {'state': "disconnected"}})
            return
        elif action == "notify_klippy_shutdown":
            self.printer.process_update({'webhooks': {'state': "shutdown"}})
            return
        elif action == "notify_klippy_ready":
            if not self.initialized:
                self.reinit_count = 0
                self.init_klipper()
                return
            self.printer.process_update({'webhooks': {'state': "ready"}})
            return
        elif action == "notify_status_update" and self.printer.state != "shutdown":
            self.printer.process_update(data)
            if 'manual_probe' in data and data['manual_probe']['is_active'] and 'zcalibrate' not in self._cur_panels:
                self.show_panel("zcalibrate")
            if ("screws_tilt_adjust" in data and "max_deviation" in data['screws_tilt_adjust']
                and not data['screws_tilt_adjust']['max_deviation']
                    and 'bed_level' not in self._cur_panels):
                self.show_panel("bed_level")
        elif action == "notify_filelist_changed":
            if self.files is not None:
                self.files.process_update(data)
            return
        elif action == "notify_metadata_update":
            self.files.request_metadata(data['filename'])
            return
        elif action == "notify_update_response":
            if 'message' in data and 'Error' in data['message']:
                logging.error(f"{action}:{data['message']}")
                self.show_popup_message(data['message'], 3, from_ws=True)
                if "KlipperScreen" in data['message']:
                    self.restart_ks()
        elif action == "notify_power_changed":
            logging.debug("Power status changed: %s", data)
            self.printer.process_power_update(data)
            self.panels['splash_screen'].check_power_status()
        elif action == "notify_gcode_response" and self.printer.state not in ["error", "shutdown"]:
            if re.match('^(?:ok\\s+)?(B|C|T\\d*):', data):
                return
            if data.startswith("// action:"):
                self.process_action(data[10:])
                return
            elif data.startswith("echo: "):
                self.show_popup_message(data[6:], 1, from_ws=True)
            elif "!! Extrude below minimum temp" in data:
                if self._cur_panels[-1] != "temperature":
                    self.show_panel("temperature", extra=self.printer.get_stat("toolhead", "extruder"))
                self.show_popup_message(_("Temperature too low to extrude"))
                return
            elif data.startswith("!! "):
                self.show_popup_message(data[3:], 3, from_ws=True)
            elif (
                "unknown" in data.lower()
                and "TESTZ" not in data
                and "MEASURE_AXES_NOISE" not in data
                and "ACCELEROMETER_QUERY" not in data
            ):
                self.show_popup_message(data, from_ws=True)
            elif "SAVE_CONFIG" in data and self.printer.state == "ready":
                script = {"script": "SAVE_CONFIG"}
                self._confirm_send_action(
                    None,
                    _("Save configuration?") + "\n\n" + _("Klipper will reboot"),
                    "printer.gcode.script",
                    script
                )
        elif action == "notify_active_spool_set":
            spool_id = data["spool_id"] if "spool_id" in data else self.spoolman_api.get_active_spool_id()
            self.update_spool_data(spool_id)
        self.process_update(action, data)

    def update_spool_data(self, spool_id=None):
        if not spool_id:
            spool_id = self.spoolman_api.get_active_spool_id()
        if not spool_id or not isinstance(spool_id, int):
            self.printer.set_active_spool(spool_id=None)
            return
        spool_data = self.spoolman_api.get_spool_details(spool_id)
        if spool_data is None:
            self.printer.set_active_spool(spool_id=spool_id)
            return
        self.printer.set_active_spool(spool_id=spool_id, spool_data=spool_data)

    def process_action(self, action):
        if action.startswith("prompt"):
            if action.startswith("prompt_begin"):
                if self.prompt is not None:
                    self.prompt.end()
                self.prompt = Prompt(self)
            if self.prompt is None:
                return
            self.prompt.decode(action)
        if action.startswith("ks_show"):
            self.parse_ks_action(action[8:].strip())

    def parse_ks_action(self, action):
        action = action.split(" ", 1)
        if len(action) == 2:
            panel, params = action
            key, value = params.split("=", 1)
            key = key.strip()
            value = value.strip()
            params = {key: ast.literal_eval(value)}
            self.show_panel(panel, **params)
        else:
            self.show_panel(*action)

    def process_update(self, *args):
        self.base_panel.process_update(*args)
        if self._cur_panels and hasattr(self.panels[self._cur_panels[-1]], "process_update"):
            self.panels[self._cur_panels[-1]].process_update(*args)

    def confirm_save(self, widget):
        buttons = [
            {"name": _("Save"), "response": Gtk.ResponseType.OK, "style": 'dialog-info'},
            {"name": _("Cancel"), "response": Gtk.ResponseType.CANCEL, "style": 'dialog-error'}
        ]
        label = Gtk.Label(label=_("Save configuration?") + "\n\n" + _("Klipper will reboot"),
                          hexpand=True, vexpand=True,
                          halign=Gtk.Align.CENTER, valign=Gtk.Align.CENTER,
                          wrap=True, wrap_mode=Pango.WrapMode.WORD_CHAR)
        grid = Gtk.Grid()
        grid.attach(label, 0, 3, 2, 1)
        offset = self.printer.get_stat("gcode_move", "homing_origin")
        zoffset = float(offset[2]) if offset else 0
        if zoffset != 0:
            sign = "+" if zoffset > 0 else "-"
            msg = f"Apply {sign}{abs(zoffset)} offset?"
            zlabel = Gtk.Label(label=msg, hexpand=True, vexpand=True, wrap=True)
            grid.attach(zlabel, 0, 1, 2, 1)
            if "Z_OFFSET_APPLY_PROBE" in self.printer.available_commands:
                apply_probe = self.gtk.Button(label=_("Save Z") + "\n" + "Probe", style="color1")
                apply_probe.set_vexpand(False)
                apply_probe.set_size_request(-1, self.gtk.dialog_buttons_height)
                apply_probe.connect("clicked", self.save, "Z_OFFSET_APPLY_PROBE")
                grid.attach(apply_probe, 0, 2, 1, 1)
            if "Z_OFFSET_APPLY_ENDSTOP" in self.printer.available_commands:
                apply_end = self.gtk.Button(label=_("Save Z") + "\n" + "Endstop", style="color2")
                apply_end.set_vexpand(False)
                apply_end.set_size_request(-1, self.gtk.dialog_buttons_height)
                apply_end.connect("clicked", self.save, "Z_OFFSET_APPLY_ENDSTOP")
                grid.attach(apply_end, 1, 2, 1, 1)
        if self.confirm is not None:
            self.gtk.remove_dialog(self.confirm)
        self.confirm = self.gtk.Dialog(
            "KlipperScreen", buttons, grid, self.save
        )

    def save(self, dialog, response_id):
        self.gtk.remove_dialog(dialog)
        if response_id == Gtk.ResponseType.OK:
            self._ws.klippy.gcode_script("SAVE_CONFIG")
        if response_id == "Z_OFFSET_APPLY_PROBE":
            self._ws.klippy.gcode_script("Z_OFFSET_APPLY_PROBE")
            self._ws.klippy.gcode_script("SAVE_CONFIG")
        if response_id == "Z_OFFSET_APPLY_ENDSTOP":
            self._ws.klippy.gcode_script("Z_OFFSET_APPLY_ENDSTOP")
            self._ws.klippy.gcode_script("SAVE_CONFIG")

    def _confirm_send_action(self, widget, text, method, params=None):
        buttons = [
            {"name": _("Accept"), "response": Gtk.ResponseType.OK, "style": 'dialog-info'},
            {"name": _("Cancel"), "response": Gtk.ResponseType.CANCEL, "style": 'dialog-error'}
        ]

        try:
            j2_temp = self.env.from_string(text)
            text = j2_temp.render()
        except Exception as e:
            logging.debug(f"Error parsing jinja for confirm_send_action\n{e}\n\n{traceback.format_exc()}")

        label = Gtk.Label(hexpand=True, vexpand=True, halign=Gtk.Align.CENTER, valign=Gtk.Align.CENTER,
                          wrap=True, wrap_mode=Pango.WrapMode.WORD_CHAR)
        label.set_markup(text)

        if self.confirm is not None:
            self.gtk.remove_dialog(self.confirm)
        self.confirm = self.gtk.Dialog(
            "KlipperScreen", buttons, label, self._confirm_send_action_response, method, params
        )

    def _confirm_send_action_response(self, dialog, response_id, method, params):
        self.gtk.remove_dialog(dialog)
        if response_id == Gtk.ResponseType.OK:
            self._send_action(None, method, params)

    def _send_action(self, widget, method, params):
        logging.info(f"{method}: {params}")
        if isinstance(widget, Gtk.Button):
            self.gtk.Button_busy(widget, True)
            self._ws.send_method(method, params, self.enable_widget, widget)
        else:
            self._ws.send_method(method, params)

    def enable_widget(self, *args):
        for x in args:
            if isinstance(x, Gtk.Button):
                GLib.timeout_add(150, self.gtk.Button_busy, x, False)

    def printer_initializing(self, msg, go_to_splash=False):
        if 'splash_screen' not in self.panels or go_to_splash:
            self.show_panel("splash_screen", remove_all=True)
        self.panels['splash_screen'].update_text(msg)
        self.log_notification(msg, 0)

    def search_power_devices(self, devices):
        found_devices = []
        if self.connected_printer is None or not devices:
            return found_devices
        devices = [str(i.strip()) for i in devices.split(',')]
        power_devices = self.printer.get_power_devices()
        if power_devices:
            found_devices = [dev for dev in devices if dev in power_devices]
            logging.info(f"Found {found_devices}", )
        return found_devices

    def power_devices(self, widget=None, devices=None, on=False):
        devs = self.search_power_devices(devices)
        for dev in devs:
            if on:
                self._ws.klippy.power_device_on(dev)
            else:
                self._ws.klippy.power_device_off(dev)

    def _init_printer(self, msg, go_to_splash=False):
        self.printer_initializing(msg, go_to_splash)
        self.initializing = False
        if self._ws.connected and not self._ws.closing:
            GLib.timeout_add_seconds(4, self.init_klipper)
        else:
            GLib.timeout_add_seconds(4, self.connect_to_moonraker)

    def connect_to_moonraker(self):
        if self.initializing:
            logging.info("Already Initializing")
            return False
        if self._ws.closing:
            logging.info("Cancelling attempt")
            return False
        self.initializing = True
        if self.reinit_count > self.max_retries or 'printer_select' in self._cur_panels:
            logging.info("Stopping Retries")
            self.printer_initializing(_("Cannot connect to Moonraker") + "\n\n"
                                      + f'{self.apiclient.status}')
            self.initializing = False
            return False
        self.server_info = self.apiclient.get_server_info()
        if not self.server_info:
            logging.info("Cannot get server info")
            if self.reinit_count > 0:
                self._init_printer(_("Cannot connect to Moonraker") + "\n\n"
                                   + _("Retrying") + f" #{self.reinit_count}")
            else:
                self._init_printer(_("Connecting to %s") % self.connecting_to_printer)
            self.initializing = False
            self.reinit_count += 1
            return False
        self._ws.initial_connect()
        return False

    def init_moonraker_components(self):
        popup = ''
        level = 2
        if self.server_info["warnings"]:
            popup += '\nMoonraker warnings:\n'
            for warning in self.server_info["warnings"]:
                warning = warning.replace('<br>', '').replace('<br/>', '\n').replace('</br>', '\n').replace(':', ':\n')
                popup += f"{warning}\n"
        if self.server_info["failed_components"]:
            popup += '\nMoonraker failed components:\n'
            for failed in self.server_info["failed_components"]:
                popup += f'[{failed}]\n'
        if self.server_info["missing_klippy_requirements"]:
            popup += '\nMissing Klipper configuration:\n'
            for missing in self.server_info["missing_klippy_requirements"]:
                popup += f'[{missing}]\n'
                level = 3
        if popup:
            self.show_popup_message(popup, level)
        if "power" in self.server_info["components"]:
            powerdevs = self.apiclient.send_request("machine/device_power/devices")
            if powerdevs is not False:
                self.printer.configure_power_devices(powerdevs)
        if "webcam" in self.server_info["components"]:
            cameras = self.apiclient.send_request("server/webcams/list")
            if cameras is not False:
                self.printer.configure_cameras(cameras['webcams'])
        if "spoolman" in self.server_info["components"]:
            self.printer.enable_spoolman()
            self.update_spool_data(self.spoolman_api.get_active_spool_id())

    def init_klipper(self):
        if self.reinit_count > self.max_retries or 'printer_select' in self._cur_panels:
            logging.info("Stopping Retries")
            return False
        if not self.server_info:
            logging.debug("Connection Lost Retrying")
            self.connect_to_moonraker()
            return False
        self.reinit_count += 1
        self.server_info = self.apiclient.get_server_info()
        logging.info(f"Moonraker info {self.server_info}")
        if self.server_info['klippy_connected'] is False:
            msg = _("Moonraker: connected") + "\n\n"
            msg += f"Klipper: {self.server_info['klippy_state']}" + "\n\n"
            if self.reinit_count <= self.max_retries:
                msg += _("Retrying") + f' #{self.reinit_count}'
            self.printer_initializing(msg)
            GLib.timeout_add_seconds(3, self.init_klipper)
            return False
        printer_info = self.apiclient.get_printer_info()
        if printer_info is False:
            self._init_printer("Unable to get printer info from moonraker")
            return False
        config = self.apiclient.send_request("printer/objects/query?configfile")
        if config is False:
            self._init_printer("Error getting printer configuration")
            return False
        self.printer.reinit(printer_info, config['status'])
        objects = self.apiclient.send_request("printer/objects/list")
        if objects and 'objects' in objects:
            self.printer.register_dynamic_sensors(objects['objects'])
        self.printer.available_commands = self.apiclient.get_gcode_help()
        info = self.apiclient.send_request("machine/system_info")
        if info and 'system_info' in info:
            self.printer.system_info = info['system_info']

        items = (
            'bed_mesh',
            'configfile',
            'display_status',
            'extruder',
            'fan',
            'gcode_move',
            'heater_bed',
            'idle_timeout',
            'pause_resume',
            'print_stats',
            'toolhead',
            'virtual_sdcard',
            'webhooks',
            'motion_report',
            'firmware_retraction',
            'exclude_object',
            'manual_probe',
            *self.printer.get_tools(),
            *self.printer.get_heaters(),
            *self.printer.get_temp_sensors(),
            *self.printer.get_fans(),
            *self.printer.get_temp_fans(),
            *self.printer.get_filament_sensors(),
            *self.printer.get_output_pins(),
            *self.printer.get_leds(),
        )

        data = self.apiclient.send_request("printer/objects/query?" + "&".join(items))
        if data is False:
            self._init_printer("Error getting printer object data")
            return False
        self.ws_subscribe()

        self.files.set_gcodes_path()
        self.power_devices(None, self._config.get_main_config().get("screen_on_devices", ""), on=True)

        logging.info("Printer initialized")
        self.initialized = True
        self.reinit_count = 0
        self.initializing = False
        self.printer.process_update(data['status'])
        self.log_notification("Printer Initialized", 1)
        return False

    def init_tempstore(self):
        if len(self.printer.get_temp_devices()) == 0:
            return False
        tempstore = self.apiclient.send_request("server/temperature_store")
        if tempstore:
            self.printer.init_temp_store(tempstore)
            if hasattr(self.panels[self._cur_panels[-1]], "update_graph_visibility"):
                self.panels[self._cur_panels[-1]].update_graph_visibility()
            if self.tempstore_timeout:
                self.remove_tempstore_timeout()
        else:
            logging.error(f'Tempstore not ready: {tempstore} Retrying in 5 seconds')
            if self.tempstore_timeout:
                return False
            if self.reinit_count < self.max_retries:
                self.reinit_count += 1
                self.tempstore_timeout = GLib.timeout_add_seconds(5, self.retry_init_tempstore)
            else:
                logging.error("Max retries reached. Stopping attempts to initialize tempstore.")
                self.remove_tempstore_timeout()
            return False

        server_config = self.apiclient.send_request("server/config")
        if server_config:
            try:
                self.printer.tempstore_size = server_config["config"]["data_store"]["temperature_store_size"]
                logging.info(f"Temperature store size: {self.printer.tempstore_size}")
            except KeyError:
                logging.error("Couldn't get the temperature store size")
        return False

    def remove_tempstore_timeout(self):
        GLib.source_remove(self.tempstore_timeout)
        self.tempstore_timeout = None
        self.reinit_count = 0

    def retry_init_tempstore(self):
        self.remove_tempstore_timeout()
        return self.init_tempstore()

    def show_keyboard(self, entry=None, event=None, box=None, close_cb=None):
        if entry is None:
            logging.debug("Error: no entry provided for keyboard")
            return
        if box is None:
            box = self.base_panel.content
        if close_cb is None:
            close_cb = self.remove_keyboard
        if self.keyboard is not None:
            self.remove_keyboard(box=box)
            entry.grab_focus()
        kbd_grid = Gtk.Grid()
        kbd_grid.set_size_request(self.gtk.content_width, self.gtk.keyboard_height)
        kbd_grid.set_vexpand(False)

        if self._config.get_main_config().getboolean("use-matchbox-keyboard", False):
            return self._show_matchbox_keyboard(kbd_grid)
        purpose = entry.get_input_purpose()
        kbd_width = 1
        if not self.vertical_mode and purpose in (Gtk.InputPurpose.DIGITS, Gtk.InputPurpose.NUMBER):
            kbd_grid.set_column_homogeneous(True)
            kbd_width = 2 if purpose == Gtk.InputPurpose.DIGITS else 3
        kbd_grid.attach(Gtk.Box(), 0, 0, 1, 1)
        kbd = self._get_keyboard(entry.get_input_purpose())
        kbd.reinit(close_cb=close_cb, entry=entry, box=box)
        kbd_grid.attach(kbd, 1, 0, kbd_width, 1)
        kbd_grid.attach(Gtk.Box(), kbd_width + 1, 0, 1, 1)
        self.keyboard = {
            "box": kbd_grid,
            "kbd": kbd
        }
        box.pack_end(kbd_grid, False, False, 0)
        box.show_all()

    def _get_keyboard(self, input_purpose):
        if input_purpose in self.keyboard_cache:
            return self.keyboard_cache[input_purpose]
        k = self.keyboard_cache[input_purpose] = Keyboard(self, lambda *args, **kwargs: None, purpose=input_purpose)
        return k

    def _show_matchbox_keyboard(self, kbd_grid):
        env = os.environ.copy()
        usrkbd = os.path.expanduser("~/.matchbox/keyboard.xml")
        if os.path.isfile(usrkbd):
            env["MB_KBD_CONFIG"] = usrkbd
        else:
            env["MB_KBD_CONFIG"] = "ks_includes/locales/keyboard.xml"
        p = subprocess.Popen(["matchbox-keyboard", "--xid"], stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE, env=env)
        xid = int(p.stdout.readline())
        logging.debug(f"XID {xid}")
        logging.debug(f"PID {p.pid}")

        keyboard = Gtk.Socket()
        kbd_grid.get_style_context().add_class("keyboard_matchbox")
        kbd_grid.attach(keyboard, 0, 0, 1, 1)
        self.base_panel.content.pack_end(kbd_grid, False, False, 0)

        self.show_all()
        keyboard.add_id(xid)

        self.keyboard = {
            "box": kbd_grid,
            "process": p,
            "socket": keyboard
        }
        return

    def remove_keyboard(self, entry=None, event=None, box=None):
        if self.keyboard is None:
            return
        if box is None:
            box = self.base_panel.content
        if 'process' in self.keyboard:
            os.kill(self.keyboard['process'].pid, SIGTERM)
        if 'kbd' in self.keyboard:
            self.keyboard['box'].remove(self.keyboard['kbd'])
        box.remove(self.keyboard['box'])
        self.keyboard = None
        if entry:
            entry.set_sensitive(False)  # Move the focus
            entry.set_sensitive(True)

    def _key_press_event(self, widget, event):
        keyval_name = Gdk.keyval_name(event.keyval)
        if keyval_name == "Escape":
            self._menu_go_back(home=True)
        elif keyval_name == "BackSpace" and len(self._cur_panels) > 1 and self.keyboard is None:
            self.base_panel.back()

    def update_size(self, *args):
        width, height = self.get_size()
        if width != self.width or height != self.height:
            logging.info(f"Size changed: {width}x{height}")
        self.width, self.height = width, height
        new_ratio = self.width / self.height
        new_mode = new_ratio < 1.0
        ratio_delta = abs(self.aspect_ratio - new_ratio)
        if ratio_delta > 0.1 and self.vertical_mode != new_mode:
            self.vertical_mode = new_mode
            self.aspect_ratio = new_ratio
            logging.info(f"Vertical mode: {self.vertical_mode}")
            self.reload_panels()


def main():
    parser = argparse.ArgumentParser(description="KlipperScreen - A GUI for Klipper")
    homedir = os.path.expanduser("~")

    parser.add_argument(
        "-c", "--configfile",
        default="", metavar='<configfile>',
        help="Location of KlipperScreen configuration file"
    )
    logdir = os.path.join(homedir, "printer_data", "logs")
    if not os.path.exists(logdir):
        logdir = "/tmp"
    parser.add_argument(
        "-l", "--logfile", default=os.path.join(logdir, "KlipperScreen.log"), metavar='<logfile>',
        help="Location of KlipperScreen logfile output"
    )
    parser.add_argument(
        "-m", "--monitor", default="0", metavar='<monitor>',
        help="Number of the monitor, that will show Klipperscreen (default: 0)"
    )
    args = parser.parse_args()

    functions.setup_logging(os.path.normpath(os.path.expanduser(args.logfile)))
    functions.patch_threading_excepthook()
    if not Gtk.init_check():
        logging.critical("Failed to initialize Gtk")
        raise RuntimeError
    try:
        win = KlipperScreen(args)
    except Exception as e:
        logging.exception(f"Failed to initialize window\n{e}\n\n{traceback.format_exc()}")
        raise RuntimeError from e
    win.connect("destroy", Gtk.main_quit)
    win.show_all()
    Gtk.main()


if __name__ == "__main__":
    try:
        main()
    except Exception as ex:
        logging.exception(f"Fatal error in main loop:\n{ex}\n\n{traceback.format_exc()}")
        os._exit(1)



import ctypes
import logging
import logging.handlers
import os
import struct
import subprocess
import sys
import threading
import traceback
from queue import SimpleQueue as Queue

dpms_loaded = False
try:
    ctypes.cdll.LoadLibrary('libXext.so.6')
    libXext = ctypes.CDLL('libXext.so.6')

    class DPMS_State:
        Fail = -1
        On = 0
        Standby = 1
        Suspend = 2
        Off = 3

    def get_DPMS_state(display_name_in_byte_string=b':0'):
        state = DPMS_State.Fail
        if not isinstance(display_name_in_byte_string, bytes):
            raise TypeError("display_name_in_byte_string must be of type bytes")

        display_name = ctypes.c_char_p(display_name_in_byte_string)
        libXext.XOpenDisplay.restype = ctypes.c_void_p
        display = ctypes.c_void_p(libXext.XOpenDisplay(display_name))

        major_opcode_p = ctypes.create_string_buffer(8)
        first_event_p = ctypes.create_string_buffer(8)

        if display.value:
            try:
                if libXext.DPMSQueryExtension(display, major_opcode_p, first_event_p) \
                        and libXext.DPMSCapable(display):
                    onoff_p = ctypes.create_string_buffer(1)
                    state_p = ctypes.create_string_buffer(2)
                    if libXext.DPMSInfo(display, state_p, onoff_p):
                        onoff = struct.unpack('B', onoff_p.raw)[0]
                        if onoff:
                            state = struct.unpack('H', state_p.raw)[0]
            finally:
                libXext.XCloseDisplay(display)
        return state

    dpms_loaded = True
except OSError as e:
    logging.error(f"Couldn't load DPMS library: {e}")
except Exception as e:
    logging.error(f"An unexpected error occurred: {e}")


def get_software_version():
    prog = ('git', '-C', os.path.dirname(__file__), 'describe', '--always', '--tags', '--long', '--dirty')
    try:
        process = subprocess.Popen(prog, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        ver, err = process.communicate()
        retcode = process.wait()
        if retcode == 0:
            version = ver.strip()
            if isinstance(version, bytes):
                version = version.decode()
            # Remove the 'g' at the start of the hash
            parts = version.split('-')
            if len(parts) > 2 and parts[-2].startswith('g'):
                parts[-2] = parts[-2][1:]  # Remove the 'g'
            version = '-'.join(parts)
            return version
        else:
            logging.debug(f"Error getting git version: {err}")
    except OSError:
        logging.exception("Error running git describe")
    return "?"


def parse_bool(value):
    return value.lower() == "true"


def patch_threading_excepthook():
    """Installs our exception handler into the threading modules Thread object
    Inspired by https://bugs.python.org/issue1230540
    """
    old_init = threading.Thread.__init__

    def new_init(self, *args, **kwargs):
        old_init(self, *args, **kwargs)
        old_run = self.run

        def run_with_excepthook(*args, **kwargs):
            try:
                old_run(*args, **kwargs)
            except (KeyboardInterrupt, SystemExit):
                raise
            except Exception:
                sys.excepthook(*sys.exc_info(), thread_identifier=threading.get_ident())

        self.run = run_with_excepthook

    threading.Thread.__init__ = new_init


# Rotating file handler based on Klipper and Moonraker's implementation
class KlipperScreenLoggingHandler(logging.handlers.RotatingFileHandler):
    def __init__(self, filename, **kwargs):
        super(KlipperScreenLoggingHandler, self).__init__(filename, **kwargs)
        self.rollover_info = {
            'header': f"{'-' * 20}KlipperScreen Log Start{'-' * 20}",
            'version': f"KlipperScreen Version: {get_software_version()}",
            'py_ver': f"Python version: {sys.version_info.major}.{sys.version_info.minor}",
        }
        self.log_start()

    def set_rollover_info(self, name, item):
        self.rollover_info[name] = item

    def doRollover(self):
        super(KlipperScreenLoggingHandler, self).doRollover()
        self.log_start()

    def log_start(self):
        for line in self.rollover_info.values():
            logging.info(line)


# Logging based on Arksine's logging setup
def setup_logging(log_file):
    root_logger = logging.getLogger()
    queue = Queue()
    queue_handler = logging.handlers.QueueHandler(queue)
    root_logger.addHandler(queue_handler)
    root_logger.setLevel(logging.DEBUG)

    stdout_hdlr = logging.StreamHandler(sys.stdout)
    stdout_fmt = logging.Formatter(
        '%(asctime)s,%(msecs)03d [%(filename)s:%(funcName)s] - %(message)s',
        '%Y%m%d %H:%M:%S')
    stdout_hdlr.setFormatter(stdout_fmt)
    fh = listener = None
    try:
        fh = KlipperScreenLoggingHandler(log_file, maxBytes=4194304, backupCount=1)
        formatter = logging.Formatter('%(asctime)s [%(filename)s:%(funcName)s()] - %(message)s')
        fh.setFormatter(formatter)
        listener = logging.handlers.QueueListener(queue, fh, stdout_hdlr)
    except Exception as e:
        print(
            f"Unable to create log file at '{os.path.normpath(log_file)}'.\n"
            f"Make sure that the folder '{os.path.dirname(log_file)}' exists\n"
            f"and KlipperScreen has Read/Write access to the folder.\n"
            f"{e}\n"
        )
    if listener is None:
        listener = logging.handlers.QueueListener(queue, stdout_hdlr)
    listener.start()

    def logging_exception_handler(ex_type, value, tb, thread_identifier=None):
        logging.exception(
            f'Uncaught exception {ex_type}: {value}\n'
            + '\n'.join([str(x) for x in [*traceback.format_tb(tb)]])
        )

    sys.excepthook = logging_exception_handler
    logging.captureWarnings(True)

    return listener, fh
